cimport cython  # NOQA
from cpython.mem cimport PyMem_Malloc, PyMem_Free
from libc.string cimport memset as c_memset
from libcpp cimport vector

import numpy
import threading

import cupy
from cupy.cuda import device
from cupy.cuda import memory
from cupy.cuda import runtime
from cupy.cuda import stream


cdef object _thread_local = threading.local()


cpdef get_current_plan():
    """Get current cuFFT plan.

    Returns:
        None or cupy.cuda.cufft.Plan1d or cupy.cuda.cufft.PlanNd
    """
    if not hasattr(_thread_local, '_current_plan'):
        _thread_local._current_plan = None
    return _thread_local._current_plan


cdef enum:
    # Actually, this is 64, but it's undocumented. For the sake
    # of safety, let us use 16, which agrees with the cuFFT doc.
    MAX_CUDA_DESCRIPTOR_GPUS = 16


cdef extern from 'cupy_cufft.h' nogil:
    # we duplicate some types here to avoid cimporting from driver/runtime,
    # as we don't include their .pxd files in the sdist
    ctypedef void* Stream 'cudaStream_t'
    ctypedef int DataType 'cudaDataType'

    ctypedef struct Complex 'cufftComplex':
        float x, y
    ctypedef struct DoubleComplex 'cufftDoubleComplex':
        double x, y

    # cuFFT Helper Function
    Result cufftCreate(Handle *plan)
    Result cufftDestroy(Handle plan)
    Result cufftSetAutoAllocation(Handle plan, int autoAllocate)
    Result cufftSetWorkArea(Handle plan, void *workArea)

    # cuFFT Stream Function
    Result cufftSetStream(Handle plan, Stream streamId)

    # cuFFT Plan Functions
    Result cufftMakePlan1d(Handle plan, int nx, Type type, int batch,
                           size_t *workSize)
    Result cufftMakePlanMany(Handle plan, int rank, int *n, int *inembed,
                             int istride, int idist, int *onembed, int ostride,
                             int odist, Type type, int batch,
                             size_t *workSize)

    # cuFFT Exec Function
    Result cufftExecC2C(Handle plan, Complex *idata, Complex *odata,
                        int direction)
    Result cufftExecR2C(Handle plan, Float *idata, Complex *odata)
    Result cufftExecC2R(Handle plan, Complex *idata, Float *odata)
    Result cufftExecZ2Z(Handle plan, DoubleComplex *idata,
                        DoubleComplex *odata, int direction)
    Result cufftExecD2Z(Handle plan, Double *idata, DoubleComplex *odata)
    Result cufftExecZ2D(Handle plan, DoubleComplex *idata, Double *odata)

    # Version
    Result cufftGetVersion(int* version)

    # cufftXt data types
    ctypedef struct XtArrayDesc 'cudaXtDesc':
        int version
        int nGPUs
        int GPUs[MAX_CUDA_DESCRIPTOR_GPUS]
        void* data[MAX_CUDA_DESCRIPTOR_GPUS]
        size_t size[MAX_CUDA_DESCRIPTOR_GPUS]
        void* cudaXtState

    ctypedef enum XtSubFormat 'cufftXtSubFormat':
        CUFFT_XT_FORMAT_INPUT
        CUFFT_XT_FORMAT_OUTPUT
        CUFFT_XT_FORMAT_INPLACE
        CUFFT_XT_FORMAT_INPLACE_SHUFFLED
        CUFFT_XT_FORMAT_1D_INPUT_SHUFFLED

    ctypedef struct XtArray 'cudaLibXtDesc':
        int version
        XtArrayDesc* descriptor
        int library
        XtSubFormat subFormat
        void* libDescriptor

    ctypedef enum XtCopyType 'cufftXtCopyType':
        CUFFT_COPY_HOST_TO_DEVICE = 0x00
        CUFFT_COPY_DEVICE_TO_HOST = 0x01
        CUFFT_COPY_DEVICE_TO_DEVICE = 0x02

    # cufftXt functions
    Result cufftXtSetGPUs(Handle plan, int nGPUs, int* gpus)
    Result cufftXtSetWorkArea(Handle plan, void** workArea)
    Result cufftXtMemcpy(Handle plan, void *dst, void *src, XtCopyType type)
    Result cufftXtExecDescriptorC2C(Handle plan, XtArray* idata,
                                    XtArray* odata, int direction)
    Result cufftXtExecDescriptorZ2Z(Handle plan, XtArray* idata,
                                    XtArray* odata, int direction)
    Result cufftXtMakePlanMany(Handle plan, int rank, long long int* n,
                               long long int* inembed,
                               long long int istride,
                               long long int idist,
                               DataType inputtype,
                               long long int* onembed,
                               long long int ostride,
                               long long int odist,
                               DataType outputtype,
                               long long int batch, size_t* workSize,
                               DataType executiontype)
    Result cufftXtExec(Handle plan, void* inarr, void* outarr, int d)


IF CUPY_CUFFT_STATIC:
    # cuFFT callback
    cdef extern from 'cupy_cufftXt.h' nogil:
        ctypedef enum callbackType 'cufftXtCallbackType':
            pass
        Result set_callback(Handle, callbackType, bint, void**)


cdef dict RESULT = {
    0: 'CUFFT_SUCCESS',
    1: 'CUFFT_INVALID_PLAN',
    2: 'CUFFT_ALLOC_FAILED',
    3: 'CUFFT_INVALID_TYPE',
    4: 'CUFFT_INVALID_VALUE',
    5: 'CUFFT_INTERNAL_ERROR',
    6: 'CUFFT_EXEC_FAILED',
    7: 'CUFFT_SETUP_FAILED',
    8: 'CUFFT_INVALID_SIZE',
    9: 'CUFFT_UNALIGNED_DATA',
    10: 'CUFFT_INCOMPLETE_PARAMETER_LIST',
    11: 'CUFFT_INVALID_DEVICE',
    12: 'CUFFT_PARSE_ERROR',
    13: 'CUFFT_NO_WORKSPACE',
    14: 'CUFFT_NOT_IMPLEMENTED',
    15: 'CUFFT_LICENSE_ERROR',
    16: 'CUFFT_NOT_SUPPORTED',
}


class CuFFTError(RuntimeError):

    def __init__(self, int result):
        self.result = result
        super(CuFFTError, self).__init__('%s' % (RESULT[result]))

    def __reduce__(self):
        return (type(self), (self.result,))


@cython.profile(False)
cpdef inline void check_result(int result) except *:
    if result != 0:
        raise CuFFTError(result)


cpdef int getVersion() except? -1:
    cdef int version, result
    result = cufftGetVersion(&version)
    check_result(result)
    return version


# This is necessary for single-batch transforms: "when batch is one, data is
# left in the GPU memory in a permutation of the natural output", see
# https://docs.nvidia.com/cuda/cufft/index.html#multiple-GPU-cufft-intermediate-helper  # NOQA
cdef _reorder_buffers(Handle plan, intptr_t xtArr, list xtArr_buffer):
    cdef int i, result, nGPUs
    cdef intptr_t temp_xtArr
    cdef XtArray* temp_arr
    cdef XtArray* arr
    cdef list gpus = []
    cdef list sizes = []
    cdef list temp_xtArr_buffer

    arr = <XtArray*>xtArr
    nGPUs = len(xtArr_buffer)
    assert nGPUs == arr.descriptor.nGPUs

    # allocate another buffer to prepare for order conversion
    for i in range(nGPUs):
        gpus.append(arr.descriptor.GPUs[i])
        sizes.append(arr.descriptor.size[i])
    temp_xtArr, temp_xtArr_buffer = _XtMalloc(gpus, sizes,
                                              CUFFT_XT_FORMAT_INPLACE)
    temp_arr = <XtArray*>temp_xtArr

    # Make a device copy to bring the data from the permuted order back to
    # the natural order. Note that this works because after FFT
    # arr.subFormat is silently changed to CUFFT_XT_FORMAT_INPLACE_SHUFFLED
    with nogil:
        result = cufftXtMemcpy(plan, <void*>temp_arr, <void*>arr,
                               CUFFT_COPY_DEVICE_TO_DEVICE)
    check_result(result)

    for i in range(nGPUs):
        # swap MemoryPointer in xtArr_buffer
        temp = temp_xtArr_buffer[i]
        temp_xtArr_buffer[i] = xtArr_buffer[i]
        xtArr_buffer[i] = temp

        # swap pointer in xtArr
        arr.descriptor.data[i] = temp_arr.descriptor.data[i]
        assert arr.descriptor.size[i] == temp_arr.descriptor.size[i]
        temp_arr.descriptor.data[i] = NULL
        temp_arr.descriptor.size[i] = 0

    # temp_xtArr now points to the old data, which is now in temp_xtArr_buffer
    # and will be deallocated after this line (out of scope)
    _XtFree(temp_xtArr)


# This is meant to replace cufftXtMalloc().
# We need to manage the buffers ourselves in order to 1. avoid excessive,
# uncessary memory usage, and 2. use CuPy's memory pool.
cdef _XtMalloc(list gpus, list sizes, XtSubFormat fmt):
    cdef XtArrayDesc* xtArr_desc
    cdef XtArray* xtArr
    cdef list xtArr_buffer = []
    cdef int i, nGPUs
    cdef size_t size

    nGPUs = len(gpus)
    assert nGPUs == len(sizes)
    xtArr_desc = <XtArrayDesc*>PyMem_Malloc(sizeof(XtArrayDesc))
    xtArr = <XtArray*>PyMem_Malloc(sizeof(XtArray))
    c_memset(xtArr_desc, 0, sizeof(XtArrayDesc))
    c_memset(xtArr, 0, sizeof(XtArray))

    xtArr_desc.nGPUs = nGPUs
    for i, (gpu, size) in enumerate(zip(gpus, sizes)):
        prev_device = runtime.getDevice()
        runtime.setDevice(gpu)
        try:
            buf = memory.alloc(size)
        finally:
            runtime.setDevice(prev_device)
        assert gpu == buf.device_id
        xtArr_buffer.append(buf)
        xtArr_desc.GPUs[i] = gpu
        xtArr_desc.data[i] = <void*><intptr_t>(buf.ptr)
        xtArr_desc.size[i] = size

    xtArr.descriptor = xtArr_desc
    xtArr.subFormat = fmt

    return <intptr_t>xtArr, xtArr_buffer


# This is meant to replace cufftXtFree().
# We only free the C structs. The underlying GPU buffers are deallocated when
# going out of scope.
cdef _XtFree(intptr_t ptr):
    cdef XtArray* xtArr = <XtArray*>ptr
    cdef XtArrayDesc* xtArr_desc = xtArr.descriptor
    PyMem_Free(xtArr_desc)
    PyMem_Free(xtArr)


cdef class Plan1d:
    def __init__(self, int nx, int fft_type, int batch, *,
                 devices=None, out=None):
        cdef Handle plan
        cdef bint use_multi_gpus = 0 if devices is None else 1
        cdef int result

        self.handle = <intptr_t>0
        self.xtArr = <intptr_t>0  # pointer to metadata for multi-GPU buffer
        self.xtArr_buffer = None  # actual multi-GPU intermediate buffer

        with nogil:
            result = cufftCreate(&plan)
            if result == 0:
                result = cufftSetAutoAllocation(plan, 0)
        check_result(result)

        self.handle = <intptr_t>plan
        self.work_area = None
        self.gpus = None

        self.gather_streams = None
        self.gather_events = None
        self.scatter_streams = None
        self.scatter_events = None

        if batch != 0:
            # set plan, work_area, gpus, streams, and events
            if not use_multi_gpus:
                self._single_gpu_get_plan(plan, nx, fft_type, batch)
            else:
                self._multi_gpu_get_plan(
                    plan, nx, fft_type, batch, devices, out)
        else:
            if use_multi_gpus:
                # multi-GPU FFT cannot transform 0-size arrays, and attempting
                # to create such a plan will error out, but we still need this
                # for bookkeeping
                if isinstance(devices, (tuple, list)):
                    self.gpus = list(devices)
                elif isinstance(devices, int) and devices > 0:
                    self.gpus = [i for i in range(int)]
                else:
                    raise ValueError

        self.nx = nx
        self.fft_type = <Type>fft_type
        self.batch = batch
        self.batch_share = None

    cdef void _single_gpu_get_plan(self, Handle plan, int nx, int fft_type,
                                   int batch) except*:
        cdef int result
        cdef size_t work_size
        cdef intptr_t ptr

        with nogil:
            result = cufftMakePlan1d(plan, nx, <Type>fft_type, batch,
                                     &work_size)

        # cufftMakePlan1d uses large memory when nx has large divisor.
        # See https://github.com/cupy/cupy/issues/1063
        if result == 2:
            cupy.get_default_memory_pool().free_all_blocks()
            with nogil:
                result = cufftMakePlan1d(plan, nx, <Type>fft_type, batch,
                                         &work_size)
        check_result(result)

        work_area = memory.alloc(work_size)
        ptr = <intptr_t>(work_area.ptr)
        with nogil:
            result = cufftSetWorkArea(plan, <void*>(ptr))
        check_result(result)

        self.work_area = work_area  # this is for cuFFT plan

    cdef void _multi_gpu_get_plan(self, Handle plan, int nx, int fft_type,
                                  int batch, devices, out) except*:
        cdef int nGPUs, min_len, result
        cdef vector.vector[int] gpus
        cdef vector.vector[size_t] work_size
        cdef list work_area = []
        cdef list gather_streams = []
        cdef list gather_events = []
        cdef vector.vector[void*] work_area_ptr

        # some sanity checks
        if runtime.is_hip:
            raise RuntimeError('hipFFT/rocFFT does not support multi-GPU FFT')
        if fft_type != CUFFT_C2C and fft_type != CUFFT_Z2Z:
            raise ValueError('Currently for multiple GPUs only C2C and Z2Z are'
                             ' supported.')
        if isinstance(devices, (tuple, list)):
            nGPUs = len(devices)
            for i in range(nGPUs):
                gpus.push_back(devices[i])
        elif isinstance(devices, int):
            nGPUs = devices
            for i in range(nGPUs):
                gpus.push_back(i)
        else:
            raise ValueError('\"devices\" should be an int or an iterable '
                             'of int.')
        if batch == 1:
            if (nx & (nx - 1)) != 0:
                raise ValueError('For multi-GPU FFT with batch = 1, the array '
                                 'size must be a power of 2.')
            if nGPUs not in (2, 4, 8, 16):
                raise ValueError('For multi-GPU FFT with batch = 1, the number'
                                 ' of devices must be 2, 4, 8, or 16.')
            if nGPUs in (2, 4):
                min_len = 64
            elif nGPUs == 8:
                min_len = 128
            else:  # nGPU = 16
                min_len = 1024
            if nx < min_len:
                raise ValueError('For {} GPUs, the array length must be at '
                                 'least {} (you have {}).'
                                 .format(nGPUs, min_len, nx))
        work_size.resize(nGPUs)

        with nogil:
            result = cufftXtSetGPUs(plan, nGPUs, gpus.data())
            if result == 0:
                result = cufftMakePlan1d(plan, nx, <Type>fft_type, batch,
                                         work_size.data())

        # cufftMakePlan1d uses large memory when nx has large divisor.
        # See https://github.com/cupy/cupy/issues/1063
        if result == 2:
            cupy.get_default_memory_pool().free_all_blocks()
            with nogil:
                result = cufftMakePlan1d(plan, nx, <Type>fft_type, batch,
                                         work_size.data())
        check_result(result)

        for i in range(nGPUs):
            prev_device = runtime.getDevice()
            runtime.setDevice(gpus[i])
            try:
                buf = memory.alloc(work_size[i])
                s = stream.Stream()
                e = stream.Event()
            finally:
                runtime.setDevice(prev_device)
            work_area.append(buf)
            work_area_ptr.push_back(<void*><intptr_t>(buf.ptr))
            gather_streams.append(s)
            gather_events.append(e)
        with nogil:
            result = cufftXtSetWorkArea(plan, work_area_ptr.data())
        check_result(result)

        self.work_area = work_area  # this is for cuFFT plan
        self.gpus = list(gpus)

        # For async, overlapped copies. We need to distinguish scatter and
        # gather because for async memcpy, the stream is on the source device
        self.gather_streams = gather_streams
        self.gather_events = gather_events
        self.scatter_streams = {}
        self.scatter_events = {}
        self._multi_gpu_get_scatter_streams_events(runtime.getDevice())

    def _multi_gpu_get_scatter_streams_events(self, int curr_device):
        '''
        create a list of streams and events on the current device
        '''
        cdef int i
        cdef list scatter_streams = []
        cdef list scatter_events = []

        assert curr_device in self.gpus
        prev_device = runtime.getDevice()
        runtime.setDevice(curr_device)
        try:
            for i in self.gpus:
                scatter_streams.append(stream.Stream())
                scatter_events.append(stream.Event())
        finally:
            runtime.setDevice(prev_device)
        self.scatter_streams[curr_device] = scatter_streams
        self.scatter_events[curr_device] = scatter_events

    def __dealloc__(self):
        cdef Handle plan = <Handle>self.handle
        cdef int dev, result

        if self.xtArr != 0:
            _XtFree(self.xtArr)
            self.xtArr = 0

        try:
            dev = runtime.getDevice()
        except Exception as e:
            # hack: the runtime module is purged at interpreter shutdown,
            # since this is not a __del__ method, we can't use
            # cupy._util.is_shutting_down()...
            return

        if plan != <Handle>0:
            with nogil:
                result = cufftDestroy(plan)
            check_result(result)
            self.handle = <intptr_t>0

        # cuFFT bug: after cufftDestroy(), the current device is mistakenly
        # set to the last device in self.gpus, so we must correct it. See
        # https://github.com/cupy/cupy/pull/2644#discussion_r347567899 and
        # NVIDIA internal ticket 2761341.
        runtime.setDevice(dev)

    def __enter__(self):
        _thread_local._current_plan = self
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        _thread_local._current_plan = None

    def fft(self, a, out, direction):
        if self.gpus is not None:
            self._multi_gpu_fft(a, out, direction)
        else:
            self._single_gpu_fft(a, out, direction)

    def _single_gpu_fft(self, a, out, direction):
        cdef intptr_t plan = self.handle
        cdef intptr_t s = stream.get_current_stream().ptr
        cdef int result

        with nogil:
            result = cufftSetStream(<Handle>plan, <Stream>s)
        check_result(result)

        if self.fft_type == CUFFT_C2C:
            execC2C(plan, a.data.ptr, out.data.ptr, direction)
        elif self.fft_type == CUFFT_R2C:
            execR2C(plan, a.data.ptr, out.data.ptr)
        elif self.fft_type == CUFFT_C2R:
            execC2R(plan, a.data.ptr, out.data.ptr)
        elif self.fft_type == CUFFT_Z2Z:
            execZ2Z(plan, a.data.ptr, out.data.ptr, direction)
        elif self.fft_type == CUFFT_D2Z:
            execD2Z(plan, a.data.ptr, out.data.ptr)
        elif self.fft_type == CUFFT_Z2D:
            execZ2D(plan, a.data.ptr, out.data.ptr)
        else:
            raise ValueError

    def _multi_gpu_setup_buffer(self, a):
        cdef XtArrayDesc* xtArr_desc
        cdef XtArray* xtArr
        cdef intptr_t ptr
        cdef list xtArr_buffer, share, sizes
        cdef int i, nGPUs
        cdef XtSubFormat fmt

        # First, get the buffers:
        # We need to manage the buffers ourselves in order to avoid excessive,
        # uncessary memory usage. Note that these buffers are used for in-place
        # transforms, and are re-used (lifetime tied to the plan).

        if isinstance(a, cupy.ndarray) or isinstance(a, numpy.ndarray):
            if self.xtArr == 0 and self.xtArr_buffer is None:
                nGPUs = len(self.gpus)

                # this is the rule for distributing the workload
                if self.batch > 1:
                    share = [self.batch // nGPUs] * nGPUs
                    for i in range(self.batch % nGPUs):
                        share[i] += 1
                else:
                    share = [1.0 / nGPUs] * nGPUs
                sizes = [int(share[i] * self.nx * a.dtype.itemsize)
                         for i in range(nGPUs)]

                # get buffer
                if isinstance(a, cupy.ndarray):
                    fmt = CUFFT_XT_FORMAT_INPLACE
                else:  # from numpy
                    fmt = CUFFT_XT_FORMAT_1D_INPUT_SHUFFLED
                ptr, xtArr_buffer = _XtMalloc(self.gpus, sizes, fmt)

                xtArr = <XtArray*>ptr
                xtArr_desc = xtArr.descriptor
                assert xtArr_desc.nGPUs == nGPUs

                self.batch_share = share
                self.xtArr = ptr
                self.xtArr_buffer = xtArr_buffer  # kept to ensure lifetime
            else:
                # After FFT the subFormat flag is silently changed to
                # CUFFT_XT_FORMAT_INPLACE_SHUFFLED. For reuse we must correct
                # it, otherwise in the next run we would encounter
                # CUFFT_INVALID_TYPE!
                ptr = self.xtArr
                xtArr = <XtArray*>ptr
                if self.batch == 1:
                    if isinstance(a, cupy.ndarray):
                        fmt = CUFFT_XT_FORMAT_INPLACE
                    else:  # from numpy
                        fmt = CUFFT_XT_FORMAT_1D_INPUT_SHUFFLED
                    xtArr.subFormat = fmt
        elif isinstance(a, list):
            # TODO(leofang): For users running Plan1d.fft() (bypassing all
            # checks in cupy.fft.fft), they are allowed to send in a list of
            # ndarrays, each of which is on a different GPU. Then, no data
            # copy is needed, just replace the pointers in the descriptor.
            raise NotImplementedError('User-managed buffer area is not yet '
                                      'supported.')
        else:
            raise ValueError('Impossible to reach.')

    def _multi_gpu_memcpy(self, a, str action):
        cdef Handle plan = <Handle>self.handle
        cdef list xtArr_buffer, share
        cdef int nGPUs, dev, s_device, start, count, result
        cdef XtArray* arr
        cdef intptr_t ptr, ptr2
        cdef size_t size

        assert isinstance(a, (cupy.ndarray, numpy.ndarray))

        start = 0
        assert a.flags.c_contiguous  # NumPy does not have _c_contiguous
        b = a.ravel()
        assert b.flags['OWNDATA'] is False
        assert self.xtArr_buffer is not None
        ptr = self.xtArr
        arr = <XtArray*>ptr
        xtArr_buffer = self.xtArr_buffer
        nGPUs = len(self.gpus)
        share = self.batch_share

        if action == 'scatter':
            if isinstance(a, cupy.ndarray):
                s_device = b.data.device_id
                if s_device not in self.scatter_streams:
                    self._multi_gpu_get_scatter_streams_events(s_device)

                # When we come here, another stream could still be
                # copying data for us, so we wait patiently...
                outer_stream = stream.get_current_stream()
                outer_stream.synchronize()

                for dev in range(nGPUs):
                    count = int(share[dev] * self.nx)
                    size = count * b.dtype.itemsize
                    curr_stream = self.scatter_streams[s_device][dev]
                    curr_event = self.scatter_events[s_device][dev]
                    xtArr_buffer[dev].copy_from_device_async(
                        b[start:start+count].data, size, curr_stream)
                    if dev != 0:
                        prev_event = self.scatter_events[s_device][dev-1]
                        curr_stream.wait_event(prev_event)
                    curr_event.record(curr_stream)
                    start += count
                assert start == b.size
                self.scatter_events[s_device][-1].synchronize()
            else:  # numpy
                ptr2 = b.ctypes.data
                with nogil:
                    result = cufftXtMemcpy(
                        plan, <void*>arr, <void*>ptr2,
                        CUFFT_COPY_HOST_TO_DEVICE)
                check_result(result)
        elif action == 'gather':
            if isinstance(a, cupy.ndarray):
                if self.batch == 1:
                    _reorder_buffers(plan, self.xtArr, xtArr_buffer)

                # When we come here, another stream could still be
                # copying data for us, so we wait patiently...
                outer_stream = stream.get_current_stream()
                outer_stream.synchronize()

                for i in range(nGPUs):
                    count = int(share[i] * self.nx)
                    size = count * b.dtype.itemsize
                    curr_stream = self.gather_streams[i]
                    curr_event = self.gather_events[i]
                    b[start:start+count].data.copy_from_device_async(
                        xtArr_buffer[i], size, curr_stream)
                    if i != 0:
                        prev_event = self.gather_events[i-1]
                        curr_stream.wait_event(prev_event)
                    curr_event.record(curr_stream)
                    start += count
                assert start == b.size
                self.gather_events[-1].synchronize()
            else:  # numpy
                ptr2 = b.ctypes.data
                with nogil:
                    result = cufftXtMemcpy(
                        plan, <void*>ptr2, <void*>arr,
                        CUFFT_COPY_DEVICE_TO_HOST)
                check_result(result)
        else:
            raise ValueError

    def _multi_gpu_fft(self, a, out, direction):
        # When we arrive here, the normal CuPy call path ensures a and out
        # reside on the same GPU -> must distribute a to all of the GPUs
        self._multi_gpu_setup_buffer(a)

        # Next, copy data to buffer
        self._multi_gpu_memcpy(a, 'scatter')

        # Actual workhorses
        # Note: mult-GPU plans cannot set stream
        cdef intptr_t plan = self.handle
        if self.fft_type == CUFFT_C2C:
            multi_gpu_execC2C(plan, self.xtArr, self.xtArr, direction)
        elif self.fft_type == CUFFT_Z2Z:
            multi_gpu_execZ2Z(plan, self.xtArr, self.xtArr, direction)
        else:
            raise ValueError

        # Gather the distributed outputs
        self._multi_gpu_memcpy(out, 'gather')

    def _output_dtype_and_shape(self, a):
        shape = list(a.shape)
        if self.fft_type == CUFFT_C2C:
            dtype = numpy.complex64
        elif self.fft_type == CUFFT_R2C:
            shape[-1] = shape[-1] // 2 + 1
            dtype = numpy.complex64
        elif self.fft_type == CUFFT_C2R:
            shape[-1] = self.nx
            dtype = numpy.float32
        elif self.fft_type == CUFFT_Z2Z:
            dtype = numpy.complex128
        elif self.fft_type == CUFFT_D2Z:
            shape[-1] = shape[-1] // 2 + 1
            dtype = numpy.complex128
        else:
            shape[-1] = self.nx
            dtype = numpy.float64
        return tuple(shape), dtype

    def get_output_array(self, a):
        shape, dtype = self._output_dtype_and_shape(a)
        return cupy.empty(shape, dtype)

    def check_output_array(self, a, out):
        """Verify shape and dtype of the output array.

        Parameters
        ----------
        a : cupy.array
            The input to the transform
        out : cupy.array
            The array where the output of the transform will be stored.
        """
        shape, dtype = self._output_dtype_and_shape(a)
        if out.shape != shape:
            raise ValueError(
                ('out must have shape {}.').format(shape))
        if out.dtype != dtype:
            raise ValueError(
                'out dtype mismatch: found {}, expected {}'.format(
                    out.dtype, dtype))


cdef class PlanNd:
    def __init__(self, object shape, object inembed, int istride,
                 int idist, object onembed, int ostride, int odist,
                 int fft_type, int batch, str order, int last_axis, last_size):
        cdef Handle plan
        cdef size_t work_size
        cdef int ndim, result
        cdef vector.vector[int] shape_arr = shape
        cdef vector.vector[int] inembed_arr
        cdef vector.vector[int] onembed_arr
        cdef int* shape_ptr = shape_arr.data()
        cdef int* inembed_ptr
        cdef int* onembed_ptr
        cdef intptr_t ptr

        self.handle = <intptr_t>0
        ndim = len(shape)

        if inembed is None:
            inembed_ptr = NULL  # ignore istride and use default strides
        else:
            inembed_arr = inembed
            inembed_ptr = inembed_arr.data()

        if onembed is None:
            onembed_ptr = NULL  # ignore ostride and use default strides
        else:
            onembed_arr = onembed
            onembed_ptr = onembed_arr.data()

        with nogil:
            result = cufftCreate(&plan)
            if result == 0:
                result = cufftSetAutoAllocation(plan, 0)
        check_result(result)

        self.handle = <intptr_t>plan
        self.gpus = None  # TODO(leofang): support multi-GPU PlanNd

        if batch == 0:
            work_size = 0
        else:
            with nogil:
                result = cufftMakePlanMany(plan, ndim, shape_ptr,
                                           inembed_ptr, istride, idist,
                                           onembed_ptr, ostride, odist,
                                           <Type>fft_type, batch,
                                           &work_size)

            # cufftMakePlanMany could use a large amount of memory
            if result == 2:
                cupy.get_default_memory_pool().free_all_blocks()
                with nogil:
                    result = cufftMakePlanMany(plan, ndim, shape_ptr,
                                               inembed_ptr, istride, idist,
                                               onembed_ptr, ostride, odist,
                                               <Type>fft_type, batch,
                                               &work_size)
            check_result(result)

        # TODO: for CUDA>=9.2 could also allow setting a work area policy
        # result = cufftXtSetWorkAreaPolicy(plan, policy, &work_size)

        work_area = memory.alloc(work_size)
        ptr = <intptr_t>(work_area.ptr)
        with nogil:
            result = cufftSetWorkArea(plan, <void*>(ptr))
        check_result(result)

        self.shape = tuple(shape)
        self.fft_type = <Type>fft_type
        self.work_area = work_area
        self.order = order  # either 'C' or 'F'
        self.last_axis = last_axis  # ignored for C2C
        self.last_size = last_size  # = None (and ignored) for C2C

    def __dealloc__(self):
        cdef Handle plan = <Handle>self.handle
        cdef int result

        if plan != <Handle>0:
            with nogil:
                result = cufftDestroy(plan)
            check_result(result)
            self.handle = <intptr_t>0

    def __enter__(self):
        _thread_local._current_plan = self
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        _thread_local._current_plan = None

    def fft(self, a, out, direction):
        cdef intptr_t plan = self.handle
        cdef intptr_t s = stream.get_current_stream().ptr
        cdef int result

        with nogil:
            result = cufftSetStream(<Handle>plan, <Stream>s)
        check_result(result)

        if self.fft_type == CUFFT_C2C:
            execC2C(plan, a.data.ptr, out.data.ptr, direction)
        elif self.fft_type == CUFFT_R2C:
            execR2C(plan, a.data.ptr, out.data.ptr)
        elif self.fft_type == CUFFT_C2R:
            execC2R(plan, a.data.ptr, out.data.ptr)
        elif self.fft_type == CUFFT_Z2Z:
            execZ2Z(plan, a.data.ptr, out.data.ptr, direction)
        elif self.fft_type == CUFFT_D2Z:
            execD2Z(plan, a.data.ptr, out.data.ptr)
        elif self.fft_type == CUFFT_Z2D:
            execZ2D(plan, a.data.ptr, out.data.ptr)
        else:
            raise ValueError

    def _output_dtype_and_shape(self, a):
        shape = list(a.shape)
        if self.fft_type == CUFFT_C2C:
            dtype = numpy.complex64
        elif self.fft_type == CUFFT_R2C:
            shape[self.last_axis] = self.last_size
            dtype = numpy.complex64
        elif self.fft_type == CUFFT_C2R:
            shape[self.last_axis] = self.last_size
            dtype = numpy.float32
        elif self.fft_type == CUFFT_Z2Z:
            dtype = numpy.complex128
        elif self.fft_type == CUFFT_D2Z:
            shape[self.last_axis] = self.last_size
            dtype = numpy.complex128
        else:  # CUFFT_Z2D
            shape[self.last_axis] = self.last_size
            dtype = numpy.float64
        return tuple(shape), dtype

    def get_output_array(self, a, order='C'):
        shape, dtype = self._output_dtype_and_shape(a)
        return cupy.empty(shape, dtype, order=order)

    def check_output_array(self, a, out):
        if out is a:
            # TODO(leofang): think about in-place transforms for C2R & R2C
            return
        if self.fft_type in (CUFFT_C2C, CUFFT_Z2Z):
            if out.shape != a.shape:
                raise ValueError('output shape mismatch')
            if out.dtype != a.dtype:
                raise ValueError('output dtype mismatch')
        else:
            if out.ndim != a.ndim:
                raise ValueError('output dimension mismatch')
            for i, size in enumerate(out.shape):
                if (i != self.last_axis and size != a.shape[i]) or \
                   (i == self.last_axis and size != self.last_size):
                    raise ValueError('output shape is incorrecct')
            if self.fft_type in (CUFFT_R2C, CUFFT_D2Z):
                if out.dtype != cupy.dtype(a.dtype.char.upper()):
                    raise ValueError('output dtype is unexpected')
            else:  # CUFFT_C2R or CUFFT_Z2D
                if out.dtype != cupy.dtype(a.dtype.char.lower()):
                    raise ValueError('output dtype is unexpected')
        if not ((out.flags.f_contiguous == a.flags.f_contiguous) and
                (out.flags.c_contiguous == a.flags.c_contiguous)):
            raise ValueError('output contiguity mismatch')


# TODO(leofang): Unify with PlanND?!
# TODO(leofang): support cufftXtSetGPUs?
cdef class XtPlanNd:
    def __init__(self, shape,
                 inembed, long long int istride, long long int idist, idtype,
                 onembed, long long int ostride, long long int odist, odtype,
                 long long int batch, edtype, *,
                 str order, int last_axis, last_size):
        # Note: we don't pass in fft_type here because it's redundant and
        # does not cover exotic types like complex32 or bf16

        cdef Handle plan
        cdef size_t work_size
        cdef int ndim, result
        cdef vector.vector[long long int] shape_arr = shape
        cdef vector.vector[long long int] inembed_arr
        cdef vector.vector[long long int] onembed_arr
        cdef long long int* shape_ptr = shape_arr.data()
        cdef long long int* inembed_ptr
        cdef long long int* onembed_ptr

        self.handle = <intptr_t>0
        ndim = len(shape)

        if inembed is None:
            inembed_ptr = NULL  # ignore istride and use default strides
        else:
            inembed_arr = inembed
            inembed_ptr = inembed_arr.data()

        if onembed is None:
            onembed_ptr = NULL  # ignore ostride and use default strides
        else:
            onembed_arr = onembed
            onembed_ptr = onembed_arr.data()

        with nogil:
            result = cufftCreate(&plan)
            if result == 0:
                result = cufftSetAutoAllocation(plan, 0)
        check_result(result)

        self.handle = <intptr_t>plan
        self.gpus = None  # TODO(leofang): support multi-GPU plans

        # determine input/output/execution types here; note that we don't
        # cimport to_cuda_dtype due to circular dependency
        from cupy._core._dtype import to_cuda_dtype
        cdef int itype = to_cuda_dtype(idtype, True)
        cdef int otype = to_cuda_dtype(odtype, True)
        cdef int etype = to_cuda_dtype(edtype, True)

        cdef long long int length
        cdef long long int full = 1
        for length in shape:
            full *= length
        length = last_size if last_size is not None else shape[-1]
        try:
            self._sanity_checks(itype, otype, etype, length, full)
        except AssertionError:
            raise ValueError('input/output/execution types mismatch')

        if batch == 0:
            work_size = 0
        else:
            with nogil:
                result = cufftXtMakePlanMany(
                    plan, ndim, shape_ptr,
                    inembed_ptr, istride, idist, <DataType>itype,
                    onembed_ptr, ostride, odist, <DataType>otype,
                    batch, &work_size, <DataType>etype)

            # cufftMakePlanMany could use a large amount of memory
            if result == 2:
                cupy.get_default_memory_pool().free_all_blocks()
                with nogil:
                    result = cufftXtMakePlanMany(
                        plan, ndim, shape_ptr,
                        inembed_ptr, istride, idist, <DataType>itype,
                        onembed_ptr, ostride, odist, <DataType>otype,
                        batch, &work_size, <DataType>etype)
            check_result(result)

        work_area = memory.alloc(work_size)
        cdef intptr_t ptr = <intptr_t>(work_area.ptr)
        with nogil:
            result = cufftSetWorkArea(plan, <void*>(ptr))
        check_result(result)

        self.shape = tuple(shape)
        self.itype = itype
        self.otype = otype
        self.etype = etype
        self.work_area = work_area
        self.order = order  # either 'C' or 'F'
        self.last_axis = last_axis  # ignored for C2C
        self.last_size = last_size  # = None (and ignored) for C2C

    def __dealloc__(self):
        cdef Handle plan = <Handle>self.handle
        cdef int result

        if plan != <Handle>0:
            with nogil:
                result = cufftDestroy(plan)
            check_result(result)
            self.handle = <intptr_t>0

    def __enter__(self):
        _thread_local._current_plan = self
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        _thread_local._current_plan = None

    def fft(self, a, out, direction):
        cdef intptr_t plan = self.handle
        cdef intptr_t s = stream.get_current_stream().ptr
        cdef int result

        with nogil:
            result = cufftSetStream(<Handle>plan, <Stream>s)
        check_result(result)
        XtExec(plan, a.data.ptr, out.data.ptr, direction)

    def _sanity_checks(self, int itype, int otype, int etype,
                       long long int last_size, long long int full_size):
        # not every possible type combination is legit
        # TODO(leofang): support bf16?
        # C2C
        if itype == runtime.CUDA_C_16F and otype == runtime.CUDA_C_16F:
            assert etype == runtime.CUDA_C_16F
        elif itype == runtime.CUDA_C_32F and otype == runtime.CUDA_C_32F:
            assert etype == runtime.CUDA_C_32F
        elif itype == runtime.CUDA_C_64F and otype == runtime.CUDA_C_64F:
            assert etype == runtime.CUDA_C_64F
        # C2R
        elif itype == runtime.CUDA_C_16F and otype == runtime.CUDA_R_16F:
            assert etype == runtime.CUDA_C_16F
        elif itype == runtime.CUDA_C_32F and otype == runtime.CUDA_R_32F:
            assert etype == runtime.CUDA_C_32F
        elif itype == runtime.CUDA_C_64F and otype == runtime.CUDA_R_64F:
            assert etype == runtime.CUDA_C_64F
        # R2C
        elif itype == runtime.CUDA_R_16F and otype == runtime.CUDA_C_16F:
            assert etype == runtime.CUDA_C_16F
        elif itype == runtime.CUDA_R_32F and otype == runtime.CUDA_C_32F:
            assert etype == runtime.CUDA_C_32F
        elif itype == runtime.CUDA_R_64F and otype == runtime.CUDA_C_64F:
            assert etype == runtime.CUDA_C_64F
        else:
            assert False

        # check fp16 runtime constraints
        # https://docs.nvidia.com/cuda/cufft/index.html#half-precision-transforms
        if etype == runtime.CUDA_C_16F:
            if int(device.get_compute_capability()) < 53:
                raise RuntimeError("this device doesn't support complex32 FFT")
            if (last_size & (last_size - 1)) != 0:
                raise ValueError('size must be power of 2')
            if full_size > 4000000000:
                raise ValueError('input array too large')
            # TODO(leofang): check if multi-GPU is requested
        # TODO(leofang): also check for bf16?
        # https://docs.nvidia.com/cuda/cufft/index.html#bfloat16-precision-transforms

    def _output_dtype_and_shape(self, a):
        shape = list(a.shape)
        if self.itype != self.otype:  # R2C or C2R
            shape[self.last_axis] = self.last_size
        if self.otype == runtime.CUDA_C_16F:
            # dtype = numpy.complex32
            raise NotImplementedError('complex32 is not supported yet, please '
                                      'allocate the output array manually')
        elif self.otype == runtime.CUDA_C_32F:
            dtype = numpy.complex64
        elif self.otype == runtime.CUDA_C_64F:
            dtype = numpy.complex128
        elif self.otype == runtime.CUDA_R_16F:
            dtype = numpy.float16
        elif self.otype == runtime.CUDA_R_32F:
            dtype = numpy.float32
        elif self.otype == runtime.CUDA_R_64F:
            dtype = numpy.float64
        return tuple(shape), dtype

    def get_output_array(self, a, order='C'):
        shape, dtype = self._output_dtype_and_shape(a)
        return cupy.empty(shape, dtype, order=order)


cpdef execC2C(intptr_t plan, intptr_t idata, intptr_t odata, int direction):
    cdef Handle h = <Handle>plan
    cdef int result

    with nogil:
        result = cufftExecC2C(h, <Complex*>idata, <Complex*>odata,
                              direction)
    check_result(result)


cpdef execR2C(intptr_t plan, intptr_t idata, intptr_t odata):
    cdef Handle h = <Handle>plan
    cdef int result

    with nogil:
        result = cufftExecR2C(h, <Float*>idata, <Complex*>odata)
    check_result(result)


cpdef execC2R(intptr_t plan, intptr_t idata, intptr_t odata):
    cdef Handle h = <Handle>plan
    cdef int result

    with nogil:
        result = cufftExecC2R(h, <Complex*>idata, <Float*>odata)
    check_result(result)


cpdef execZ2Z(intptr_t plan, intptr_t idata, intptr_t odata, int direction):
    cdef Handle h = <Handle>plan
    cdef int result

    with nogil:
        result = cufftExecZ2Z(h, <DoubleComplex*>idata,
                              <DoubleComplex*>odata, direction)
    check_result(result)


cpdef execD2Z(intptr_t plan, intptr_t idata, intptr_t odata):
    cdef Handle h = <Handle>plan
    cdef int result

    with nogil:
        result = cufftExecD2Z(h, <Double*>idata, <DoubleComplex*>odata)
    check_result(result)


cpdef execZ2D(intptr_t plan, intptr_t idata, intptr_t odata):
    cdef Handle h = <Handle>plan
    cdef int result

    with nogil:
        result = cufftExecZ2D(h, <DoubleComplex*>idata, <Double*>odata)
    check_result(result)


cpdef multi_gpu_execC2C(intptr_t plan, intptr_t idata, intptr_t odata,
                        int direction):
    cdef Handle h = <Handle>plan
    cdef int result

    with nogil:
        result = cufftXtExecDescriptorC2C(h, <XtArray*>idata,
                                          <XtArray*>odata, direction)
    check_result(result)


cpdef multi_gpu_execZ2Z(intptr_t plan, intptr_t idata, intptr_t odata,
                        int direction):
    cdef Handle h = <Handle>plan
    cdef int result

    with nogil:
        result = cufftXtExecDescriptorZ2Z(h, <XtArray*>idata,
                                          <XtArray*>odata, direction)
    check_result(result)


cpdef XtExec(intptr_t plan, intptr_t idata, intptr_t odata, int direction):
    cdef Handle h = <Handle>plan
    cdef int result

    with nogil:
        result = cufftXtExec(h, <void*>idata, <void*>odata, direction)
    check_result(result)


cpdef intptr_t setCallback(
        intptr_t plan, int cb_type, bint is_load, intptr_t aux_arr=0):
    cdef Handle h = <Handle>plan  # no-cython-lint
    cdef int result  # no-cython-lint
    cdef void** callerInfo  # no-cython-lint

    IF CUPY_CUFFT_STATIC:
        with nogil:
            if aux_arr > 0:
                callerInfo = (<void**>(&aux_arr))
            else:
                callerInfo = NULL
            result = set_callback(
                h, <callbackType>cb_type, is_load, callerInfo)
        check_result(result)
    ELSE:
        raise RuntimeError('cuFFT is dynamically linked and thus does not '
                           'support callback')



import numpy
import pytest

import cupy
from cupy import testing
import cupyx.scipy.special


class _TestBase:

    @testing.with_requires('scipy<1.16')
    def test_ndtr(self):
        self.check_unary_linspace0_1('ndtr')

    @testing.with_requires('scipy<1.16')
    def test_log_ndtr(self):
        self.check_unary_linspace0_1('log_ndtr')

    def test_ndtri(self):
        self.check_unary_linspace0_1('ndtri')

    @testing.with_requires("scipy<1.14")
    def test_logit(self):
        self.check_unary_lower_precision('logit')

    @testing.with_requires("scipy<1.14")
    def test_expit(self):
        self.check_unary_lower_precision('expit')

    @testing.with_requires("scipy>=1.8.0rc0", "scipy<1.14")
    def test_log_expit(self):
        self.check_unary_lower_precision('log_expit')


atol = {'default': 1e-14, cupy.float64: 1e-14}
rtol = {'default': 1e-5, cupy.float64: 1e-14}

# not all functions pass at the stricter tolerances above
atol_low = {'default': 5e-4, cupy.float64: 1e-12}
rtol_low = {'default': 5e-4, cupy.float64: 1e-12}


@testing.with_requires('scipy')
class TestSpecial(_TestBase):

    def _check_unary(self, a, name, scp):
        import scipy.special  # NOQA

        return getattr(scp.special, name)(a)

    @testing.for_dtypes(['e', 'f', 'd'])
    @testing.numpy_cupy_allclose(atol=atol, rtol=rtol, scipy_name='scp')
    def check_unary(self, name, xp, scp, dtype):
        a = xp.linspace(-10, 10, 100, dtype=dtype)
        return self._check_unary(a, name, scp)

    @testing.for_dtypes(['e', 'f', 'd'])
    @testing.numpy_cupy_allclose(atol=atol_low, rtol=rtol_low,
                                 scipy_name='scp')
    def check_unary_lower_precision(self, name, xp, scp, dtype):
        a = xp.linspace(-10, 10, 100, dtype=dtype)
        return self._check_unary(a, name, scp)

    @testing.for_dtypes(['e', 'f', 'd'])
    @testing.numpy_cupy_allclose(atol=atol, rtol=rtol, scipy_name='scp')
    def check_unary_linspace0_1(self, name, xp, scp, dtype):
        p = xp.linspace(0, 1, 1000, dtype=dtype)
        return self._check_unary(p, name, scp)

    def test_logit_nonfinite(self):
        logit = cupyx.scipy.special.logit
        assert float(logit(0)) == -numpy.inf
        assert float(logit(1)) == numpy.inf
        assert numpy.isnan(float(logit(1.1)))
        assert numpy.isnan(float(logit(-0.1)))

    @pytest.mark.parametrize('inverse', [False, True],
                             ids=['boxcox', 'inv_boxcox'])
    @testing.for_dtypes(['e', 'f', 'd'])
    @testing.numpy_cupy_allclose(atol=atol, rtol=rtol, scipy_name='scp')
    def test_boxcox(self, xp, scp, dtype, inverse):
        import scipy.special  # NOQA

        # outputs are only finite over range (0, 1)
        x = xp.linspace(0.001, 1000, 1000, dtype=dtype).reshape((1, 1000))
        lmbda = xp.asarray([-5, 0, 5], dtype=dtype).reshape((3, 1))
        result = scp.special.boxcox(x, lmbda)
        if inverse:
            result = scp.special.inv_boxcox(result, lmbda)
        return result

    def test_boxcox_nonfinite(self):
        boxcox = cupyx.scipy.special.boxcox
        assert float(boxcox(0, -5)) == -numpy.inf
        assert numpy.isnan(float(boxcox(-0.1, 5)))

    @pytest.mark.parametrize('inverse', [False, True],
                             ids=['boxcox', 'inv_boxcox'])
    @testing.for_dtypes(['e', 'f', 'd'])
    @testing.numpy_cupy_allclose(atol=atol, rtol=rtol, scipy_name='scp')
    def test_boxcox1p(self, xp, scp, dtype, inverse):
        import scipy.special  # NOQA

        x = xp.linspace(-0.99, 1000, 1000, dtype=dtype).reshape((1, 1000))
        lmbda = xp.asarray([-5, 0, 5], dtype=dtype).reshape((3, 1))
        result = scp.special.boxcox1p(x, lmbda)
        if inverse:
            result = scp.special.inv_boxcox1p(result, lmbda)
        return result

    def test_boxcox1p_nonfinite(self):
        boxcox1p = cupyx.scipy.special.boxcox1p
        assert float(boxcox1p(-1, -5)) == -numpy.inf
        assert numpy.isnan(float(boxcox1p(-1.1, 5)))


@testing.with_requires('scipy')
class TestFusionSpecial(_TestBase):

    def _check_unary(self, a, name, scp):
        import scipy.special  # NOQA

        @cupy.fuse()
        def f(x):
            return getattr(scp.special, name)(x)

        return f(a)

    @testing.for_dtypes(['e', 'f', 'd'])
    @testing.numpy_cupy_allclose(atol=atol, rtol=rtol, scipy_name='scp')
    def check_unary(self, name, xp, scp, dtype):
        a = testing.shaped_arange((2, 3), xp, dtype)
        return self._check_unary(a, name, scp)

    @testing.for_dtypes(['e', 'f', 'd'])
    @testing.numpy_cupy_allclose(atol=atol_low, rtol=rtol_low,
                                 scipy_name='scp')
    def check_unary_lower_precision(self, name, xp, scp, dtype):
        a = testing.shaped_arange((2, 3), xp, dtype)
        return self._check_unary(a, name, scp)

    @testing.for_dtypes(['e', 'f', 'd'])
    @testing.numpy_cupy_allclose(atol=atol, rtol=rtol, scipy_name='scp')
    def check_unary_linspace0_1(self, name, xp, scp, dtype):
        a = xp.linspace(0, 1, 1000, dtype)
        return self._check_unary(a, name, scp)


class _TestDistributionsBase:

    def _test_scalar(self, function, args, expected, rtol=1e-12, atol=1e-12):
        special = cupyx.scipy.special
        function = getattr(special, function)
        testing.assert_allclose(function(*args), expected, rtol=rtol,
                                atol=atol)


@testing.with_requires('scipy')
class TestTwoArgumentDistribution(_TestDistributionsBase):

    @testing.with_requires('scipy<1.16')
    @pytest.mark.skipif(cupy.cuda.runtime.is_hip,
                        reason="avoid failures observed on HIP")
    @pytest.mark.parametrize('function', ['chdtr', 'chdtrc', 'chdtri',
                                          'pdtr', 'pdtrc', 'pdtri'])
    @testing.for_float_dtypes()
    @testing.numpy_cupy_allclose(atol=1e-5, rtol=1e-5, scipy_name='scp')
    def test_linspace_broadcast(self, xp, scp, dtype, function):
        import scipy.special  # NOQA

        func = getattr(scp.special, function)
        # chdtr* comparisons fail at < 1 degree of freedom
        minval = 1 if function.startswith('chdtr') else -1
        v = xp.arange(minval, 10, dtype=dtype)[:, xp.newaxis]
        if function in ['chdtri', 'pdtri']:
            # concentrate values around probability range ([0, 1])
            x = xp.linspace(-.1, 1.3, 20, dtype=dtype)
        else:
            # concentrate mostly on valid, positive values
            x = xp.linspace(-1, 10, 20, dtype=dtype)
        return func(v, x[xp.newaxis, :])

    @testing.for_float_dtypes()
    @testing.for_int_dtypes(name='int_dtype', no_bool=True)
    @testing.numpy_cupy_allclose(atol=1e-5, rtol=1e-5, scipy_name='scp')
    def test_linspace_pdtri(self, xp, scp, int_dtype, dtype):
        import scipy.special  # NOQA

        func = getattr(scp.special, 'pdtri')
        k = xp.arange(1, 10, dtype=int_dtype)[:, xp.newaxis]
        y = xp.linspace(0, 1, 20, dtype=dtype)[xp.newaxis, :]
        return func(k, y)

    @pytest.mark.parametrize(
        'function, args, expected',
        [('chdtr', (1, 0), 0.0),
         ('chdtr', (0.7, cupy.inf), 1.0),
         ('chdtr', (0.6, 3), 0.957890536704110),
         ('chdtrc', (1, 0), 1.0),
         ('chdtrc', (0.6, 3), 1 - 0.957890536704110),
         ('chdtri', (1, 1), 0.0),
         ('chdtri', (0.6, 1 - 0.957890536704110), 3),
         ]
    )
    def test_scalar(self, function, args, expected):
        self._test_scalar(function, args, expected)


@testing.with_requires('scipy')
class TestThreeArgumentDistributions(_TestDistributionsBase):

    @pytest.mark.parametrize('function', ['btdtr', 'btdtri', 'fdtr', 'fdtrc',
                                          'fdtri', 'gdtr', 'gdtrc'])
    @testing.for_float_dtypes()
    @testing.numpy_cupy_allclose(atol=1e-5, rtol=1e-5, scipy_name='scp')
    def test_linspace_broadcasted(self, xp, scp, dtype, function):
        """Linspace with three arguments.

        This method uses first two arguments with mostly non-negative values.
        In some cases, the last argument is constrained to range [0, 1]
        """
        if (
            function in ['btdtr', 'btdtri'] and
            testing.installed('scipy>=1.12.0rc1')
        ):
            pytest.skip('btdtr and btdtri are deprecated since SciPy 1.12')

        import scipy.special  # NOQA

        func = getattr(scp.special, function)
        # a and b should be positive
        a = xp.linspace(-1, 21, 30, dtype=dtype)[:, xp.newaxis, xp.newaxis]
        b = xp.linspace(-1, 21, 30, dtype=dtype)[xp.newaxis, :, xp.newaxis]
        if function in ['fdtri', 'btdtr', 'btdtri']:
            # x should be in [0, 1] so concentrate values around that
            x = xp.linspace(-0.1, 1.3, 20, dtype=dtype)
        else:
            # x should be non-negative, but test with at least 1 negative value
            x = xp.linspace(-1, 10, 20, dtype=dtype)
        x = x[xp.newaxis, xp.newaxis, :]
        return func(a, b, x)

    # omit test with scipy < 1.5 due to change in ufunc type signatures
    @pytest.mark.parametrize('function', ['bdtr', 'bdtrc', 'bdtri'])
    @testing.for_float_dtypes()
    @testing.for_signed_dtypes(name='int_dtype')
    @testing.numpy_cupy_allclose(atol=1e-5, rtol=1e-5, scipy_name='scp')
    @testing.with_requires('scipy>=1.5.0')
    @pytest.mark.skipif(cupy.cuda.runtime.is_hip,
                        reason="avoid failures observed on HIP")
    def test_binomdist_linspace(self, xp, scp, function, dtype, int_dtype):
        import scipy.special  # NOQA

        # Skip cases deprecated in SciPy 1.5+ via this Cython code:
        # https://github.com/scipy/scipy/blob/cdb9b034d46c7ba0cacf65a9b2848c5d49c286c4/scipy/special/_legacy.pxd#L39-L43  # NOQA
        # All type casts except `dld->d` should raise a DeprecationWawrning
        # However on the SciPy side, this shows up as a SystemError
        #    SystemError: <class 'DeprecationWarning'> returned a result with an exception set  # NOQA
        safe_cast = xp.result_type(int_dtype, 'l') == xp.dtype('l')
        safe_cast &= xp.result_type(int_dtype, dtype) == xp.float64
        if not safe_cast:
            return xp.zeros((1,))

        func = getattr(scp.special, function)
        n = xp.linspace(0, 80, 80, dtype=int_dtype)[xp.newaxis, :, xp.newaxis]

        # broadcast to create k <= n
        k = xp.linspace(0, 1, 10, dtype=dtype)
        k = k[:, xp.newaxis, xp.newaxis] * n
        p = xp.linspace(0, 1, 5, dtype=dtype)[xp.newaxis, xp.newaxis, :]
        return func(k, n, p)

    @pytest.mark.parametrize('function', ['nbdtr', 'nbdtrc', 'nbdtri'])
    @testing.for_float_dtypes()
    @testing.for_signed_dtypes(name='int_dtype')
    @testing.numpy_cupy_allclose(atol=1e-5, rtol=1e-5, scipy_name='scp')
    def test_negbinomdist_linspace(self, xp, scp, function, dtype, int_dtype):
        import scipy.special  # NOQA

        func = getattr(scp.special, function)
        n = xp.linspace(0, 20, 20, dtype=int_dtype)[xp.newaxis, :, xp.newaxis]
        k = xp.linspace(0, 20, 20, dtype=int_dtype)[:, xp.newaxis, xp.newaxis]
        p = xp.linspace(0, 1, 5, dtype=dtype)[xp.newaxis, xp.newaxis, :]
        return func(k, n, p)

    @pytest.mark.parametrize(
        'function, args, expected',
        [('btdtr', (1, 1, 1), 1.0),
         ('btdtri', (1, 1, 1), 1.0),
         ('betainc', (1, 1, 0), 0.0),
         # Computed using Wolfram Alpha: CDF[FRatioDistribution[1e-6, 5], 10]
         ('fdtr', (1e-6, 5, 10), 0.9999940790193488),
         ('fdtrc', (1, 1, 0), 1.0),
         # Computed using Wolfram Alpha:
         #   1 - CDF[FRatioDistribution[2, 1/10], 1e10]
         ('fdtrc', (2, 0.1, 1e10), 0.2722378462129351),
         # From Wolfram Alpha:
         #   CDF[FRatioDistribution[1/10, 1], 3] = 0.8756751669632106...
         ('fdtri', (0.1, 1, 0.8756751669632106), 3.0),
         ('gdtr', (1, 1, 0), 0.0),
         ('gdtr', (1, 1, cupy.inf), 1.0),
         ('gdtrc', (1, 1, 0), 1.0),
         ('bdtr', (1, 1, 0.5), 1.0),
         ('bdtrc', (1, 3, 0.5), 0.5),
         ('bdtri', (1, 3, 0.5), 0.5),
         ('nbdtr', (1, 1, 1), 1.0),
         ('nbdtrc', (1, 1, 1), 0.0),
         ('nbdtri', (1, 1, 1), 1.0),
         ]
    )
    def test_scalar(self, function, args, expected):
        self._test_scalar(function, args, expected)



from cupy._core._carray cimport shape_t
from cupy._core cimport _kernel
from cupy._core cimport _optimize_config
from cupy._core cimport _reduction
from cupy._core cimport _scalar
from cupy._core.core cimport compile_with_cache
from cupy._core.core cimport _ndarray_base
from cupy._core.core cimport _internal_ascontiguousarray
from cupy._core cimport internal
from cupy.cuda cimport cub
from cupy.cuda cimport function
from cupy.cuda cimport memory
from cupy_backends.cuda.api cimport runtime

import math
import string
from cupy import _environment
from cupy._core._kernel import _get_param_info
from cupy.cuda import driver
from cupy import _util


cdef function.Function _create_cub_reduction_function(
        name, block_size, items_per_thread,
        reduce_type, params, arginfos, identity,
        pre_map_expr, reduce_expr, post_map_expr,
        _kernel._TypeMap type_map, preamble, options):
    # A (incomplete) list of internal variables:
    # _J            : the index of an element in the array
    # ROCm5.3 and above requires c++14
    if runtime._is_hip_environment:
        options += ('--std=c++14',)
    else:
        # 1. static_assert needs at least C++11 in NVRTC
        # 2. starting CUDA 12.2, fp16/bf16 headers are intertwined, but due to
        #    license issue we can't yet bundle bf16 headers. CUB offers us a
        #    band-aid solution to avoid including the latter (NVIDIA/cub#478,
        #    nvbugs 3641496).
        # 3. Recent CCCL versions need C++17.
        options += ('--std=c++17', '-DCUB_DISABLE_BF16_SUPPORT')

    cdef str backend
    if runtime._is_hip_environment:
        # In ROCm, we need to set the include path. This does not work for
        # hiprtc as of ROCm 3.5.0, so we must use hipcc.
        options += ('-I' + _rocm_path + '/include', '-O2')
        backend = 'nvcc'  # this is confusing...
        jitify = False
    else:
        # use nvrtc
        backend = 'nvrtc'
        # We rely on the type traits in cccl to avoid using jitify
        jitify = False

    # TODO(leofang): try splitting the for-loop into full tiles and partial
    # tiles to utilize LoadDirectBlockedVectorized? See, for example,
    # https://github.com/NVlabs/cub/blob/c3cceac115c072fb63df1836ff46d8c60d9eb304/cub/agent/agent_reduce.cuh#L311-L346

    cdef str module_code = _get_cub_header_include()
    module_code += '''
${type_preamble}
${preamble}

typedef ${reduce_type} _type_reduce;

static_assert(sizeof(_type_reduce) <= 32,
    "The intermediate reduction type is assumed to be at most 32 bytes.");

// Compile-time constants for CUB template specializations
#define ITEMS_PER_THREAD  ${items_per_thread}
#define BLOCK_SIZE        ${block_size}

// for hipCUB: use the hipcub namespace
#ifdef __HIP_DEVICE_COMPILE__
#define cub hipcub
#endif

#if defined FIRST_PASS
    typedef type_in0_raw  type_mid_in;
    typedef _type_reduce  type_mid_out;
    #define POST_MAP(a)   out0 = a;
#elif defined SECOND_PASS
    typedef _type_reduce  type_mid_in;
    typedef type_out0_raw type_mid_out;
    #define POST_MAP(a)   (${post_map_expr})
#else  // one-pass reduction
    typedef type_in0_raw  type_mid_in;
    typedef type_out0_raw type_mid_out;
    #define POST_MAP(a)   (${post_map_expr})
#endif

struct _reduction_op {
    __device__ __forceinline__ _type_reduce operator()(
        const _type_reduce &a, const _type_reduce &b) const {
        return ${reduce_expr};
    }
};

extern "C"
__global__ void ${name}(${params}) {
  unsigned int _tid = threadIdx.x;
'''

    if pre_map_expr == 'in0':
        module_code += '''
  // Specialize BlockLoad type for faster (?) loading
  typedef cub::BlockLoad<_type_reduce, BLOCK_SIZE,
                         ITEMS_PER_THREAD, cub::BLOCK_LOAD_DIRECT> BlockLoadT;

  // Shared memory for loading
  __shared__ typename BlockLoadT::TempStorage temp_storage_load;
'''

    module_code += '''
  // Specialize BlockReduce type for our thread block
  typedef cub::BlockReduce<_type_reduce, BLOCK_SIZE> BlockReduceT;

  // Shared memory for reduction
  __shared__ typename BlockReduceT::TempStorage temp_storage;

  // Declare reduction operation
  _reduction_op op;

  // input & output raw pointers
  const type_mid_in* _in0 = static_cast<const type_mid_in*>(_raw_in0);
  type_mid_out* _out0 = static_cast<type_mid_out*>(_raw_out0);

  // Per-thread tile data
  _type_reduce _sdata[ITEMS_PER_THREAD];
  #pragma unroll
  for (int j = 0; j < ITEMS_PER_THREAD; j++) {
      _sdata[j] = _type_reduce(${identity});
  }

  // each block handles the reduction of 1 segment
  size_t segment_idx = size_t(blockIdx.x) * _segment_size;
  const type_mid_in* segment_head = _in0 + segment_idx;
  size_t i = 0;  // tile head within the segment
  int tile_size = (BLOCK_SIZE * ITEMS_PER_THREAD < _segment_size ?
                   BLOCK_SIZE * ITEMS_PER_THREAD :
                   _segment_size);
  sizeT _seg_size = _segment_size;

  #if defined FIRST_PASS
  // for two-pass reduction only: "last segment" is special
  if (_array_size > 0) {
      if (_array_size - segment_idx <= _segment_size) {
          _seg_size = _array_size - segment_idx;
      }
      #ifdef __HIP_DEVICE_COMPILE__
      // We don't understand HIP...
      __syncthreads();  // Propagate the new value back to memory
      #endif
  }
  #endif

  // loop over tiles within 1 segment
  _type_reduce aggregate = _type_reduce(${identity});
  for (i = 0; i < _seg_size; i += BLOCK_SIZE * ITEMS_PER_THREAD) {
      // for the last tile
      if (_seg_size - i <= tile_size) { tile_size = _seg_size - i; }
'''

    if pre_map_expr == 'in0':
        module_code += '''
      // load a tile
      BlockLoadT(temp_storage_load).Load(segment_head + i, _sdata, tile_size,
                                         _type_reduce(${identity}));
'''
    else:  # pre_map_expr could be something like "in0 != type_in0_raw(0)"
        module_code += '''
      // load a tile
      #pragma unroll
      for (int j = 0; j < ITEMS_PER_THREAD; j++) {
          // index of the element in a tile
          int e_idx = _tid * ITEMS_PER_THREAD + j;

          // some pre_map_expr uses _J internally...
          #if defined FIRST_PASS
          IndexT _J = (segment_idx + i + e_idx);
          #else  // only one pass
          IndexT _J = (segment_idx + i + e_idx) % _seg_size;
          #endif

          if (e_idx < tile_size) {
              const type_mid_in in0 = *(segment_head + i + e_idx);
              _sdata[j] = static_cast<_type_reduce>(${pre_map_expr});
          } else {
              _sdata[j] = _type_reduce(${identity});
          }
      }
'''

    module_code += '''
      // Compute block reduction
      // Note that the output is only meaningful for thread 0
      aggregate = op(aggregate, BlockReduceT(temp_storage).Reduce(_sdata, op));

      __syncthreads();  // for reusing temp_storage
  }

  if (_tid == 0) {
      type_mid_out& out0 = *(_out0 + blockIdx.x);
      POST_MAP(aggregate);
  }
}
'''

    module_code = string.Template(module_code).substitute(
        name=name,
        block_size=block_size,
        items_per_thread=items_per_thread,
        reduce_type=reduce_type,
        params=_get_cub_kernel_params(params, arginfos),
        identity=identity,
        reduce_expr=reduce_expr,
        pre_map_expr=pre_map_expr,
        post_map_expr=post_map_expr,
        type_preamble=type_map.get_typedef_code(),
        preamble=preamble)

    # To specify the backend, we have to explicitly spell out the default
    # values for arch, cachd, prepend_cupy_headers, ... to bypass cdef/cpdef
    # limitation...
    module = compile_with_cache(
        module_code, options, arch=None, cachd_dir=None,
        prepend_cupy_headers=True, backend=backend, translate_cucomplex=False,
        enable_cooperative_groups=False, name_expressions=None,
        log_stream=None, jitify=jitify)
    return module.get_function(name)


@_util.memoize(for_each_device=True)
def _SimpleCubReductionKernel_get_cached_function(
        map_expr, reduce_expr, post_map_expr, reduce_type,
        params, arginfos, _kernel._TypeMap type_map,
        name, block_size, identity, preamble,
        options, cub_params):
    items_per_thread = cub_params[0]
    name = name.replace('cupy_', 'cupy_cub_')
    name = name.replace('cupyx_', 'cupyx_cub_')
    return _create_cub_reduction_function(
        name, block_size, items_per_thread,
        reduce_type, params, arginfos, identity,
        map_expr, reduce_expr, post_map_expr,
        type_map, preamble, options)


cdef str _cub_path = _environment.get_cub_path()
cdef str _nvcc_path = _environment.get_nvcc_path()
cdef str _rocm_path = _environment.get_rocm_path()
cdef str _hipcc_path = _environment.get_hipcc_path()
cdef str _cub_header = None


cdef str _get_cub_header_include():
    global _cub_header
    if _cub_header is not None:
        return _cub_header

    assert _cub_path is not None
    if _cub_path == '<bundle>':
        _cub_header = '''
#include <cub/block/block_reduce.cuh>
#include <cub/block/block_load.cuh>
'''
    elif _cub_path == '<ROCm>':
        # As of ROCm 3.5.0, the block headers cannot be included by themselves
        # (many macros left undefined), so we must use the master header.
        _cub_header = '''
#include <hipcub/hipcub.hpp>
'''
    return _cub_header


# make it cpdef'd for unit tests
cpdef inline tuple _can_use_cub_block_reduction(
        list in_args, list out_args, tuple reduce_axis, tuple out_axis):
    '''
    If CUB BlockReduce can be used, this function returns a tuple of the needed
    parameters, otherwise returns None.
    '''
    cdef tuple axis_permutes_cub
    cdef _ndarray_base in_arr
    cdef Py_ssize_t contiguous_size = 1
    cdef str order

    # detect whether CUB headers exists somewhere:
    if _cub_path is None:
        import warnings
        warnings.warn('CUB headers are not found.', RuntimeWarning)
        return None

    # we currently support reductions with 1 input and 1 output
    if len(in_args) != 1 or len(out_args) != 1:
        return None

    in_arr = in_args[0]

    # the axes might not be sorted when we arrive here...
    reduce_axis = tuple(sorted(reduce_axis))
    out_axis = tuple(sorted(out_axis))

    # check reduction axes, if not contiguous then fall back to old kernel
    if in_arr._f_contiguous:
        order = 'F'
        if not cub._cub_device_segmented_reduce_axis_compatible(
                reduce_axis, in_arr.ndim, order):
            return None
        axis_permutes_cub = reduce_axis + out_axis
    elif in_arr._c_contiguous:
        order = 'C'
        if not cub._cub_device_segmented_reduce_axis_compatible(
                reduce_axis, in_arr.ndim, order):
            return None
        axis_permutes_cub = out_axis + reduce_axis
    else:
        return None
    if axis_permutes_cub != tuple(range(in_arr.ndim)):
        return None

    # full-reduction of N-D array: need to invoke the kernel twice
    cdef bint full_reduction = True if len(out_axis) == 0 else False

    # check if the number of elements is too large
    # (ref: cupy/cupy#3309 for CUB limit)
    for i in reduce_axis:
        contiguous_size *= in_arr.shape[i]
    if contiguous_size > 0x7fffffffffffffff or contiguous_size == 0:
        return None
    if full_reduction:
        # assume a GPU has at most 64 GB of physical memory
        if contiguous_size > 0x1000000000:
            return None
    else:
        # the number of blocks to be launched exceeds INT_MAX:
        if in_arr.size // contiguous_size > 0x7fffffff:
            return None

    # rare event (mainly for conda-forge users): nvcc is not found!
    if not runtime._is_hip_environment:
        if _nvcc_path is None:
            return None
    else:
        if _hipcc_path is None:
            return None

    return (axis_permutes_cub, contiguous_size, full_reduction)


# similar to cupy._core._kernel._get_kernel_params()
cdef str _get_cub_kernel_params(tuple params, tuple arginfos):
    cdef _kernel.ParameterInfo p
    cdef _kernel._ArgInfo arginfo
    cdef lst = []
    cdef str c_type, c_name
    cdef int i
    assert len(params) == len(arginfos)

    for i, (p, arginfo) in enumerate(zip(params, arginfos)):
        c_name = arginfo.get_c_var_name(p)
        if i < len(params) - 2:
            c_type = 'const void*' if p.is_const else 'void*'
        else:
            # for segment size and array size
            c_type = arginfo.get_param_c_type(p)
        lst.append('{} {}'.format(c_type, c_name))
    return ', '.join(lst)


cdef Py_ssize_t _cub_default_block_size = (
    256 if runtime._is_hip_environment else 512)


cdef (Py_ssize_t, Py_ssize_t) _get_cub_block_specs(  # NOQA
        Py_ssize_t contiguous_size):
    # This is recommended in the CUB internal and should be an
    # even number
    items_per_thread = 4

    # Calculate the reduction block dimensions.
    # Ideally, we want each block to handle one segment, so:
    # 1. block size < segment size: the block loops over the segment
    # 2. block size >= segment size: the segment fits in the block
    block_size = (contiguous_size + items_per_thread - 1) // items_per_thread
    block_size = internal.clp2(block_size)
    warp_size = 32 if not runtime._is_hip_environment else 64
    if block_size < warp_size:
        block_size = warp_size
    elif block_size > _cub_default_block_size:
        block_size = _cub_default_block_size

    return items_per_thread, block_size


cdef _scalar.CScalar _cub_convert_to_c_scalar(
        Py_ssize_t segment_size, Py_ssize_t value):
    if segment_size > 0x7fffffff:
        return _scalar.scalar_to_c_scalar(value)
    else:
        return _scalar.CScalar.from_int32(value)


cdef inline void _cub_two_pass_launch(
        str name, Py_ssize_t block_size, Py_ssize_t segment_size,
        Py_ssize_t items_per_thread, str reduce_type, tuple params,
        list in_args, list out_args,
        str identity, str pre_map_expr, str reduce_expr, str post_map_expr,
        _kernel._TypeMap type_map, str preamble,
        tuple options, stream) except*:
    '''
    Notes:
    1. Two-pass reduction: the first pass distributes an even share over
       a number of blocks (with block_size threads), and the second pass
       does reduction over 1 block of threads
    '''

    cdef list out_args_2nd_pass = [out_args[0]]
    cdef Py_ssize_t contiguous_size, out_block_num
    cdef function.Function func
    cdef memory.MemoryPointer memptr
    cdef str post_map_expr1, post_map_expr2, f
    cdef list inout_args
    cdef tuple cub_params
    cdef size_t gridx, blockx
    cdef _ndarray_base in_arr

    # fair share
    contiguous_size = min(segment_size, block_size * items_per_thread)
    out_block_num = (segment_size + contiguous_size - 1) // contiguous_size
    assert out_block_num <= 0x7fffffff

    # Because we can't know sizeof(reduce_type) in advance, here we
    # conservatively assume it's 32 bytes and allocate a work area
    memptr = memory.alloc(out_block_num * 32)
    out_args[0] = memptr

    # ************************ 1st pass ************************
    name += '_pass1'
    inout_args = [in_args[0], out_args[0],
                  _cub_convert_to_c_scalar(segment_size, contiguous_size),
                  _cub_convert_to_c_scalar(segment_size, segment_size)]
    cub_params = (items_per_thread,)

    if 'mean' in name:
        post_map_expr1 = post_map_expr.replace('_in_ind.size()', '1.0')
        post_map_expr1 = post_map_expr1.replace('_out_ind.size()', '1.0')
    elif any((f in name for f in ('argmax', 'argmin'))):
        # Workaround: in NumPy the indices are always generated based on
        # a C-order array (since PyArray_ContiguousFromAny was called).
        # We have to do a conversion here (?) since we do not retain the
        # info on strides.
        # TODO(leofang): improve this workaround
        in_arr = in_args[0]
        if in_arr.ndim > 1 and in_arr._f_contiguous:
            in_arr = _internal_ascontiguousarray(in_arr)
            inout_args[0] = in_args[0] = in_arr
        post_map_expr1 = post_map_expr
    else:
        post_map_expr1 = post_map_expr

    # Retrieve the kernel function
    func = _SimpleCubReductionKernel_get_cached_function(
        pre_map_expr, reduce_expr, post_map_expr1, reduce_type,
        params,
        _kernel._get_arginfos(inout_args),
        type_map,
        name, block_size, identity, preamble,
        ('-DFIRST_PASS=1',), cub_params)

    # Kernel arguments passed to the __global__ function.
    gridx = <size_t>(out_block_num * block_size)
    blockx = <size_t>block_size

    # Launch the kernel
    func.linear_launch(gridx, inout_args, 0, blockx, stream)

    # ************************ 2nd pass ************************
    name = name[:-1] + '2'
    contiguous_size = out_block_num
    out_block_num = 1
    in_args = out_args
    out_args = out_args_2nd_pass
    inout_args = [in_args[0], out_args[0],
                  _cub_convert_to_c_scalar(segment_size, contiguous_size),
                  _cub_convert_to_c_scalar(segment_size, segment_size)]

    # For mean()
    if 'mean' in name:
        post_map_expr2 = post_map_expr.replace('_in_ind.size()',
                                               '_array_size')
        post_map_expr2 = post_map_expr2.replace('_out_ind.size()', '1.0')
    else:
        post_map_expr2 = post_map_expr

    # Retrieve the kernel function
    func = _SimpleCubReductionKernel_get_cached_function(
        'in0', reduce_expr, post_map_expr2, reduce_type,
        params,
        _kernel._get_arginfos(inout_args),
        type_map,
        name, block_size, identity, preamble,
        ('-DSECOND_PASS=1',), cub_params)

    # Kernel arguments passed to the __global__ function.
    gridx = <size_t>(out_block_num * block_size)
    blockx = <size_t>block_size

    # Launch the kernel
    func.linear_launch(gridx, inout_args, 0, blockx, stream)


cdef inline void _launch_cub(
        self, out_block_num, block_size, block_stride,
        in_args, out_args, in_shape, out_shape, type_map,
        map_expr, reduce_expr, post_map_expr, reduce_type,
        stream, params, cub_params) except *:
    cdef bint full_reduction
    cdef Py_ssize_t contiguous_size, items_per_thread
    cdef function.Function func

    # Kernel arguments passed to the __global__ function.
    items_per_thread = cub_params[0]
    contiguous_size = cub_params[1]
    full_reduction = cub_params[2]

    if full_reduction:
        _cub_two_pass_launch(
            self.name, block_size, contiguous_size, items_per_thread,
            reduce_type, params, in_args, out_args, self.identity,
            map_expr, reduce_expr, post_map_expr,
            type_map, self.preamble, (), stream)
        return
    else:
        inout_args = (
            in_args + out_args +
            [_cub_convert_to_c_scalar(
                contiguous_size, contiguous_size),
             _cub_convert_to_c_scalar(
                 contiguous_size, 0)])
        arginfos = _kernel._get_arginfos(inout_args)
        func = _SimpleCubReductionKernel_get_cached_function(
            map_expr, reduce_expr, post_map_expr, reduce_type,
            params, arginfos, type_map,
            self.name, block_size, self.identity, self.preamble,
            (), cub_params)

        func.linear_launch(
            out_block_num * block_size, inout_args, 0, block_size, stream)


def _get_cub_optimized_params(
        self, optimize_config, in_args, out_args, in_shape, out_shape,
        type_map, map_expr, reduce_expr, post_map_expr, reduce_type,
        stream, full_reduction, out_block_num, contiguous_size, params):
    in_args = [_reduction._optimizer_copy_arg(a) for a in in_args]
    out_args = [_reduction._optimizer_copy_arg(a) for a in out_args]

    items_per_thread, block_size = (
        _get_cub_block_specs(contiguous_size))
    default_block_size_log = math.floor(math.log2(block_size))
    default_items_per_thread = items_per_thread

    def target_func(block_size, items_per_thread):
        block_stride = block_size * items_per_thread
        cub_params = (
            items_per_thread, contiguous_size, full_reduction)
        _launch_cub(
            self,
            out_block_num, block_size, block_stride, in_args, out_args,
            in_shape, out_shape, type_map, map_expr, reduce_expr,
            post_map_expr, reduce_type, stream, params, cub_params)

    def suggest_func(trial):
        block_size_log = trial.suggest_int('block_size_log', 5, 10)
        block_size = 2 ** block_size_log
        items_per_thread = trial.suggest_int(
            'items_per_thread', 2, 32, step=2)

        trial.set_user_attr('block_size', block_size)
        return block_size, items_per_thread

    # CUDA_ERROR_LAUNCH_OUT_OF_RESOURCES is a possible error
    optimize_impl = optimize_config.optimize_impl
    best = optimize_impl(
        optimize_config, target_func, suggest_func,
        default_best={
            'block_size_log': default_block_size_log,
            'items_per_thread': default_items_per_thread,
        }, ignore_error=(driver.CUDADriverError,))

    return best.params['items_per_thread'], best.user_attrs['block_size']


cdef bint _try_to_call_cub_reduction(
        self, list in_args, list out_args, const shape_t& a_shape,
        stream, optimize_context, tuple key,
        map_expr, reduce_expr, post_map_expr,
        reduce_type, _kernel._TypeMap type_map,
        tuple reduce_axis, tuple out_axis, const shape_t& out_shape,
        _ndarray_base ret) except *:
    """Try to use cub.

    Updates `ret` and returns a boolean value whether cub is used.

    Note: input_expr and output_expr are not used in CUB kernels.
    """
    cdef tuple axis_permutes
    cdef tuple params, opt_params
    cdef shape_t in_shape
    cdef Py_ssize_t i
    cdef Py_ssize_t contiguous_size = -1
    cdef Py_ssize_t block_size, block_stride, out_block_num = 0

    # decide to use CUB or not
    can_use_cub = _can_use_cub_block_reduction(
        in_args, out_args, reduce_axis, out_axis)

    if can_use_cub is None:
        return False

    axis_permutes, contiguous_size, full_reduction = can_use_cub

    in_shape = _reduction._set_permuted_args(
        in_args, axis_permutes, a_shape, self.in_params)

    if in_args[0]._f_contiguous:
        ret._set_contiguous_strides(ret.dtype.itemsize, False)
        out_args[0] = ret

    if not full_reduction:  # just need one pass
        out_block_num = 1  # = number of segments
        for i in out_axis:
            out_block_num *= in_shape[i]

        if 'mean' in self.name:
            post_map_expr = post_map_expr.replace(
                '_in_ind.size()', '_segment_size')
            post_map_expr = post_map_expr.replace(
                '_out_ind.size()', '1.0')

    if contiguous_size > 0x7fffffff:  # INT_MAX
        size_type = 'uint64'
    else:
        size_type = 'int32'
    type_map = _kernel._TypeMap(type_map._pairs + (('sizeT', size_type),))
    params = (self._params[0:2]
              + _get_param_info(size_type + ' _segment_size', True)
              + _get_param_info(size_type + ' _array_size', True))

    # HACK for ReductionKernel:
    # 1. input/output arguments might not be named as in0/out0
    # 2. pre-/post- maps might not contain in0/out0
    # 3. type_map does not contain the expected names (type_in0_raw and
    #    type_out0_raw)
    cdef str old_in0 = params[0].name, old_out0 = params[1].name
    if old_in0 != 'in0' or old_out0 != 'out0':
        # avoid overwriting self's attributes
        params = (_get_param_info('T in0', True)
                  + _get_param_info('T out0', False)
                  + params[2:])
        map_expr = map_expr.replace(old_in0, 'in0')
        post_map_expr = post_map_expr.replace(old_out0, 'out0')
        type_map = _kernel._TypeMap(type_map._pairs + (
            ('type_in0_raw', in_args[0].dtype.type),
            ('type_out0_raw', out_args[0].dtype.type),
        ))

    # Calculate the reduction block dimensions.
    optimize_context = _optimize_config.get_current_context()
    if optimize_context is None:
        # Calculate manually
        items_per_thread, block_size = _get_cub_block_specs(contiguous_size)
    else:
        # Optimize dynamically
        key = ('cub_reduction',) + key
        opt_params = optimize_context.get_params(key)
        if opt_params is None:
            opt_params = _get_cub_optimized_params(
                self,
                optimize_context.config, in_args, out_args,
                in_shape, out_shape, type_map, map_expr, reduce_expr,
                post_map_expr, reduce_type, stream,
                full_reduction, out_block_num, contiguous_size, params)
            optimize_context.set_params(key, opt_params)
        items_per_thread, block_size = opt_params

    block_stride = block_size * items_per_thread
    cub_params = (items_per_thread, contiguous_size, full_reduction)

    _launch_cub(
        self,
        out_block_num,
        block_size,
        block_stride,
        in_args, out_args,
        in_shape, out_shape,
        type_map,
        map_expr, reduce_expr, post_map_expr, reduce_type,
        stream, params, cub_params)

    return True



from cupy.cuda cimport function
from cupy._core cimport internal


cdef class CArray(function.CPointer):

    cdef void init(
            self, void* data_ptr, Py_ssize_t data_size,
            const shape_t& shape, const strides_t& strides) except*:
        cdef size_t ndim = shape.size()
        assert ndim == strides.size()
        assert ndim <= MAX_NDIM
        cdef Py_ssize_t* shape_and_strides = (
            self.val.shape_and_strides)
        cdef size_t i

        self.val.data = data_ptr
        self.val.size = data_size
        for i in range(ndim):
            shape_and_strides[i] = shape[i]
            shape_and_strides[i + ndim] = strides[i]
        self.ptr = <void*>&self.val


cdef class CIndexer(function.CPointer):

    cdef void init(self, Py_ssize_t size, const shape_t &shape) except*:
        cdef size_t ndim = shape.size()
        assert ndim <= MAX_NDIM
        self.val.size = size
        cdef Py_ssize_t i
        for i in range(<Py_ssize_t>shape.size()):
            self.val.shape_and_index[i] = shape[i]
        self.ptr = <void*>&self.val


cdef class Indexer:

    cdef void init(self, const shape_t& shape):
        self.shape = shape
        self.size = internal.prod(shape)
        self._index_32_bits = self.size <= (1 << 31)

    @property
    def ndim(self):
        return self.shape.size()

    cdef function.CPointer get_pointer(self):
        cdef CIndexer indexer = CIndexer.__new__(CIndexer)
        indexer.init(self.size, self.shape)
        return indexer


cdef inline Indexer _indexer_init(const shape_t& shape):
    cdef Indexer indexer = Indexer.__new__(Indexer)
    indexer.init(shape)
    return indexer



# Bessel Functions
from cupyx.scipy.special._bessel import i0  # NOQA
from cupyx.scipy.special._bessel import i0e  # NOQA
from cupyx.scipy.special._bessel import i1  # NOQA
from cupyx.scipy.special._bessel import i1e  # NOQA
from cupyx.scipy.special._bessel import j0  # NOQA
from cupyx.scipy.special._bessel import j1  # NOQA
from cupyx.scipy.special._bessel import k0  # NOQA
from cupyx.scipy.special._bessel import k0e  # NOQA
from cupyx.scipy.special._bessel import k1  # NOQA
from cupyx.scipy.special._bessel import k1e  # NOQA
from cupyx.scipy.special._bessel import y0  # NOQA
from cupyx.scipy.special._bessel import y1  # NOQA
from cupyx.scipy.special._bessel import yn  # NOQA
from cupyx.scipy.special._spherical_bessel import spherical_yn  # NOQA

# Raw statistical functions
from cupyx.scipy.special._stats_distributions import bdtr  # NOQA
from cupyx.scipy.special._stats_distributions import bdtrc  # NOQA
from cupyx.scipy.special._stats_distributions import bdtri  # NOQA
from cupyx.scipy.special._stats_distributions import btdtr  # NOQA
from cupyx.scipy.special._stats_distributions import btdtri  # NOQA
from cupyx.scipy.special._stats_distributions import fdtr  # NOQA
from cupyx.scipy.special._stats_distributions import fdtrc  # NOQA
from cupyx.scipy.special._stats_distributions import fdtri  # NOQA
from cupyx.scipy.special._stats_distributions import gdtr  # NOQA
from cupyx.scipy.special._stats_distributions import gdtrc  # NOQA
from cupyx.scipy.special._stats_distributions import nbdtr  # NOQA
from cupyx.scipy.special._stats_distributions import nbdtrc  # NOQA
from cupyx.scipy.special._stats_distributions import nbdtri  # NOQA
from cupyx.scipy.special._stats_distributions import pdtr  # NOQA
from cupyx.scipy.special._stats_distributions import pdtrc  # NOQA
from cupyx.scipy.special._stats_distributions import pdtri  # NOQA
from cupyx.scipy.special._stats_distributions import chdtr  # NOQA
from cupyx.scipy.special._stats_distributions import chdtrc  # NOQA
from cupyx.scipy.special._stats_distributions import chdtri  # NOQA
from cupyx.scipy.special._stats_distributions import ndtr  # NOQA
from cupyx.scipy.special._stats_distributions import log_ndtr  # NOQA
from cupyx.scipy.special._stats_distributions import ndtri  # NOQA
from cupyx.scipy.special._statistics import logit  # NOQA
from cupyx.scipy.special._statistics import expit  # NOQA
from cupyx.scipy.special._statistics import log_expit  # NOQA
from cupyx.scipy.special._statistics import boxcox  # NOQA
from cupyx.scipy.special._statistics import boxcox1p  # NOQA
from cupyx.scipy.special._statistics import inv_boxcox  # NOQA
from cupyx.scipy.special._statistics import inv_boxcox1p  # NOQA

# Information Theory functions
from cupyx.scipy.special._convex_analysis import entr  # NOQA
from cupyx.scipy.special._convex_analysis import huber  # NOQA
from cupyx.scipy.special._convex_analysis import kl_div  # NOQA
from cupyx.scipy.special._convex_analysis import pseudo_huber  # NOQA
from cupyx.scipy.special._convex_analysis import rel_entr  # NOQA

# Gamma and related functions
from cupyx.scipy.special._gamma import gamma  # NOQA
from cupyx.scipy.special._gammaln import gammaln  # NOQA
from cupyx.scipy.special._loggamma import loggamma  # NOQA
from cupyx.scipy.special._gammasgn import gammasgn  # NOQA
from cupyx.scipy.special._gammainc import gammainc  # NOQA
from cupyx.scipy.special._gammainc import gammaincinv  # NOQA
from cupyx.scipy.special._gammainc import gammaincc  # NOQA
from cupyx.scipy.special._gammainc import gammainccinv  # NOQA
from cupyx.scipy.special._beta import beta  # NOQA
from cupyx.scipy.special._beta import betaln  # NOQA
from cupyx.scipy.special._beta import betainc  # NOQA
from cupyx.scipy.special._beta import betaincinv  # NOQA
from cupyx.scipy.special._digamma import digamma as psi  # NOQA
from cupyx.scipy.special._gamma import rgamma  # NOQA
from cupyx.scipy.special._polygamma import polygamma  # NOQA
from cupyx.scipy.special._gammaln import multigammaln  # NOQA
from cupyx.scipy.special._digamma import digamma  # NOQA
from cupyx.scipy.special._poch import poch  # NOQA

# Error function and Fresnel integrals
from cupyx.scipy.special._erf import erf  # NOQA
from cupyx.scipy.special._erf import erfc  # NOQA
from cupyx.scipy.special._erf import erfcx  # NOQA
from cupyx.scipy.special._erf import erfinv  # NOQA
from cupyx.scipy.special._erf import erfcinv  # NOQA

# Legendre functions
from cupyx.scipy.special._lpmv import lpmv  # NOQA
from cupyx.scipy.special._sph_harm import sph_harm  # NOQA

# Other special functions
from cupyx.scipy.special._binom import binom  # NOQA
from cupyx.scipy.special._exp1 import exp1  # NOQA
from cupyx.scipy.special._expi import expi  # NOQA
from cupyx.scipy.special._expn import expn  # NOQA
from cupyx.scipy.special._softmax import softmax  # NOQA
from cupyx.scipy.special._logsoftmax import log_softmax  # NOQA
from cupyx.scipy.special._zeta import zeta  # NOQA
from cupyx.scipy.special._zetac import zetac  # NOQA

# Convenience functions
from cupyx.scipy.special._basic import cbrt  # NOQA
from cupyx.scipy.special._basic import exp10  # NOQA
from cupyx.scipy.special._basic import exp2  # NOQA
from cupyx.scipy.special._basic import radian  # NOQA
from cupyx.scipy.special._basic import cosdg  # NOQA
from cupyx.scipy.special._basic import sindg  # NOQA
from cupyx.scipy.special._basic import tandg  # NOQA
from cupyx.scipy.special._basic import cotdg  # NOQA
from cupyx.scipy.special._basic import log1p  # NOQA
from cupyx.scipy.special._basic import expm1  # NOQA
from cupyx.scipy.special._basic import exprel  # NOQA
from cupyx.scipy.special._basic import cosm1  # NOQA
from cupy._math.rounding import round  # NOQA
from cupyx.scipy.special._xlogy import xlogy  # NOQA
from cupyx.scipy.special._xlogy import xlog1py  # NOQA
from cupyx.scipy.special._logsumexp import logsumexp  # NOQA
from cupy._math.special import sinc  # NOQA

# Elliptic functions
from cupyx.scipy.special._ellip import ellipk  # NOQA
from cupyx.scipy.special._ellip import ellipkm1  # NOQA
from cupyx.scipy.special._ellip import ellipj  # NOQA



import unittest
import pytest

import cupy
from cupy import testing
from cupy_tests.core_tests.fusion_tests import fusion_utils


@testing.slow
@pytest.mark.skipif(
    cupy.cuda.runtime.is_hip, reason='HIP does not support this')
class TestFusionExample(unittest.TestCase):
    def generate_inputs(self, xp):
        shape = (8, 64, 112, 112)
        _, chan, _, _ = shape
        x = testing.shaped_random(shape, xp, 'float32', scale=10, seed=0)
        gamma = xp.ones(chan)
        beta = xp.zeros(chan)
        running_mean = xp.zeros(chan)
        running_var = xp.ones(chan)
        size = x.size // gamma.size
        adjust = size / max(size - 1., 1.)
        return (x, gamma, beta, running_mean, running_var, size, adjust), {}

    @fusion_utils.check_fusion()
    def test_batchnorm(self, xp):
        def batchnorm(x, gamma, beta, running_mean, running_var, size, adjust):
            decay = 0.9
            eps = 2e-5
            expander = (None, slice(None), None, None)

            gamma = gamma[expander]
            beta = beta[expander]
            mean = xp.sum(x, axis=(0, 2, 3)) / size
            diff = x - mean[expander]
            var = xp.sum(diff * diff, axis=(0, 2, 3)) / size
            inv_std = 1. / xp.sqrt(var + eps)
            y = gamma * diff * inv_std[expander] + beta

            running_mean *= decay
            running_mean += (1 - decay) * mean
            running_var *= decay
            running_var += (1 - decay) * adjust * var
            return y

        return batchnorm



import numpy as np
from numpy import linalg
from cupyx.scipy.linalg import bandwidth
from pytest import raises
from cupy import testing


class TestBadwidth:
    @testing.for_all_dtypes(no_complex=True)
    def test_bandwidth_non2d_input(self, dtype):
        A = np.array([1, 2, 3], dtype=dtype)
        raises(linalg.LinAlgError, bandwidth, A)
        A = np.array([[[1, 2, 3], [4, 5, 6]]], dtype=dtype)
        raises(linalg.LinAlgError, bandwidth, A)

    @testing.for_all_dtypes(no_complex=True)
    def test_bandwidth_square_symmetric_inputs(self, dtype):
        n = 10
        for k in range(1, 10):
            R = np.zeros([n, n], dtype=dtype)
            # form a banded matrix inplace
            R[[x for x in range(n)], [x for x in range(n)]] = 1
            R[[x for x in range(n - k)], [x for x in range(k, n)]] = 1
            R[[x for x in range(1, n)], [x for x in range(n - 1)]] = 1
            R[[x for x in range(k, n)], [x for x in range(n - k)]] = 1
            testing.assert_array_equal(bandwidth(R), (k, k))

    @testing.for_all_dtypes(no_complex=True)
    def test_bandwidth_square_asymmetric_inputs_c(self, dtype):
        n = 20
        a = 5
        b = 4
        R = np.zeros([n, n], dtype=dtype)
        # form a banded matrix inplace
        R[[x for x in range(n)], [x for x in range(n)]] = 1
        R[[x for x in range(a, n)], [x for x in range(n - a)]] = 1
        R[[x for x in range(n - b)], [x for x in range(b, n)]] = 1
        testing.assert_array_equal(bandwidth(R), (a, b))

    @testing.for_all_dtypes(no_complex=True)
    def test_bandwidth_square_asymmetric_inputs_f(self, dtype):
        n = 20
        a = 5
        b = 4
        R = np.zeros([n, n], dtype=dtype, order='F')
        # form a banded matrix inplace
        R[[x for x in range(n)], [x for x in range(n)]] = 1
        R[[x for x in range(a, n)], [x for x in range(n - a)]] = 1
        R[[x for x in range(n - b)], [x for x in range(b, n)]] = 1
        testing.assert_array_equal(bandwidth(R), (a, b))

    @testing.for_all_dtypes(no_complex=True)
    def test_bandwidth_square_inputs_f(self, dtype):
        n = 10
        for k in range(1, 10):
            R = np.zeros([n, n], dtype=dtype, order='F')
            # form a banded matrix inplace
            R[[x for x in range(n)], [x for x in range(n)]] = 1
            R[[x for x in range(n - k)], [x for x in range(k, n)]] = 1
            R[[x for x in range(1, n)], [x for x in range(n - 1)]] = 1
            R[[x for x in range(k, n)], [x for x in range(n - k)]] = 1
            testing.assert_array_equal(bandwidth(R), (k, k))

    @testing.for_all_dtypes(no_complex=True)
    def test_bandwidth_rect_inputs_c(self, dtype):
        n, m = 10, 20
        k = 5
        R = np.zeros([n, m], dtype=dtype)
        # form a banded matrix inplace
        R[[x for x in range(n)], [x for x in range(n)]] = 1
        R[[x for x in range(n - k)], [x for x in range(k, n)]] = 1
        R[[x for x in range(1, n)], [x for x in range(n - 1)]] = 1
        R[[x for x in range(k, n)], [x for x in range(n - k)]] = 1
        testing.assert_array_equal(bandwidth(R), (k, k))

    @testing.for_all_dtypes(no_complex=True)
    def test_bandwidth_rect_inputs_f(self, dtype):
        n, m = 10, 20
        k = 5
        R = np.zeros([n, m], dtype=dtype, order='F')
        # form a banded matrix inplace
        R[[x for x in range(n)], [x for x in range(n)]] = 1
        R[[x for x in range(n - k)], [x for x in range(k, n)]] = 1
        R[[x for x in range(1, n)], [x for x in range(n - 1)]] = 1
        R[[x for x in range(k, n)], [x for x in range(n - k)]] = 1
        testing.assert_array_equal(bandwidth(R), (k, k))



from libcpp cimport vector
from cupy.cuda cimport memory

from cupy.cuda.function cimport CPointer
from cupy.cuda.function cimport Module
from cupy._core._carray cimport shape_t
from cupy._core._carray cimport strides_t


cdef class _ndarray_base:
    cdef:
        object __weakref__
        readonly Py_ssize_t size
        public shape_t _shape
        public strides_t _strides
        readonly bint _c_contiguous
        readonly bint _f_contiguous
        # To do fast indexing in the CArray class
        readonly bint _index_32_bits
        readonly object dtype
        readonly memory.MemoryPointer data
        # TODO(niboshi): Return arbitrary owner object as `base` if the
        # underlying memory is UnownedMemory.
        readonly _ndarray_base base

    cdef _init_fast(self, const shape_t& shape, dtype, bint c_order)
    cdef inline bint is_host_accessible(self) except*
    cpdef item(self)
    cpdef tolist(self)
    cpdef bytes tobytes(self, order=*)
    cpdef tofile(self, fid, sep=*, format=*)
    cpdef dump(self, file)
    cpdef bytes dumps(self)
    cpdef _ndarray_base astype(
        self, dtype, order=*, casting=*, subok=*, copy=*)
    cpdef _ndarray_base copy(self, order=*)
    cpdef _ndarray_base view(self, dtype=*, array_class=*)
    cpdef fill(self, value)
    cpdef _ndarray_base swapaxes(self, Py_ssize_t axis1, Py_ssize_t axis2)
    cpdef _ndarray_base flatten(self, order=*)
    cpdef _ndarray_base ravel(self, order=*)
    cpdef _ndarray_base squeeze(self, axis=*)
    cpdef _ndarray_base take(self, indices, axis=*, out=*)
    cpdef put(self, indices, values, mode=*)
    cpdef repeat(self, repeats, axis=*)
    cpdef choose(self, choices, out=*, mode=*)
    cpdef sort(self, int axis=*)
    cpdef _ndarray_base argsort(self, axis=*)
    cpdef partition(self, kth, int axis=*)
    cpdef _ndarray_base argpartition(self, kth, axis=*)
    cpdef tuple nonzero(self)
    cpdef _ndarray_base compress(self, condition, axis=*, out=*)
    cpdef _ndarray_base diagonal(self, offset=*, axis1=*, axis2=*)
    cpdef _ndarray_base max(self, axis=*, out=*, keepdims=*)
    cpdef _ndarray_base argmax(self, axis=*, out=*, dtype=*, keepdims=*)
    cpdef _ndarray_base min(self, axis=*, out=*, keepdims=*)
    cpdef _ndarray_base argmin(self, axis=*, out=*, dtype=*, keepdims=*)
    cpdef _ndarray_base ptp(self, axis=*, out=*, keepdims=*)
    cpdef _ndarray_base clip(self, min=*, max=*, out=*)
    cpdef _ndarray_base round(self, decimals=*, out=*)

    cpdef _ndarray_base trace(self, offset=*, axis1=*, axis2=*, dtype=*, out=*)
    cpdef _ndarray_base sum(self, axis=*, dtype=*, out=*, keepdims=*)
    cpdef _ndarray_base cumsum(self, axis=*, dtype=*, out=*)
    cpdef _ndarray_base mean(self, axis=*, dtype=*, out=*, keepdims=*)
    cpdef _ndarray_base var(self, axis=*, dtype=*, out=*, ddof=*, keepdims=*)
    cpdef _ndarray_base std(self, axis=*, dtype=*, out=*, ddof=*, keepdims=*)
    cpdef _ndarray_base prod(self, axis=*, dtype=*, out=*, keepdims=*)
    cpdef _ndarray_base cumprod(self, axis=*, dtype=*, out=*)
    cpdef _ndarray_base _add_reduceat(self, indices, axis, dtype, out)
    cpdef _ndarray_base all(self, axis=*, out=*, keepdims=*)
    cpdef _ndarray_base any(self, axis=*, out=*, keepdims=*)
    cpdef _ndarray_base conj(self)
    cpdef _ndarray_base conjugate(self)
    cpdef get(self, stream=*, order=*, out=*, blocking=*)
    cpdef set(self, arr, stream=*)
    cpdef _ndarray_base reduced_view(self, dtype=*)
    cpdef _update_c_contiguity(self)
    cpdef _update_f_contiguity(self)
    cpdef _update_contiguity(self)
    cpdef _set_shape_and_strides(self, const shape_t& shape,
                                 const strides_t& strides,
                                 bint update_c_contiguity,
                                 bint update_f_contiguity)
    cdef _ndarray_base _view(self, subtype, const shape_t& shape,
                             const strides_t& strides,
                             bint update_c_contiguity,
                             bint update_f_contiguity, obj)
    cpdef _set_contiguous_strides(
        self, Py_ssize_t itemsize, bint is_c_contiguous)
    cdef CPointer get_pointer(self)
    cpdef object toDlpack(self)


cpdef _ndarray_base _internal_ascontiguousarray(_ndarray_base a)
cpdef _ndarray_base _internal_asfortranarray(_ndarray_base a)
cpdef _ndarray_base ascontiguousarray(_ndarray_base a, dtype=*)
cpdef _ndarray_base asfortranarray(_ndarray_base a, dtype=*)

cpdef Module compile_with_cache(str source, tuple options=*, arch=*,
                                cachd_dir=*, prepend_cupy_headers=*,
                                backend=*, translate_cucomplex=*,
                                enable_cooperative_groups=*,
                                name_expressions=*, log_stream=*,
                                bint jitify=*)


# TODO(niboshi): Move to _routines_creation.pyx
cpdef _ndarray_base array(
    obj, dtype=*, bint copy=*, order=*, bint subok=*, Py_ssize_t ndmin=*,
    bint blocking=*)
cpdef _ndarray_base _convert_object_with_cuda_array_interface(a)

cdef _ndarray_base _ndarray_init(subtype, const shape_t& shape, dtype, obj)

cdef _ndarray_base _create_ndarray_from_shape_strides(
    subtype, const shape_t& shape, const strides_t& strides, dtype, obj)



import unittest
import warnings

import numpy
import pytest

import cupy
from cupy import testing
from cupy.exceptions import ComplexWarning


class TestFlatiter(unittest.TestCase):

    def test_base(self):
        for xp in (numpy, cupy):
            a = xp.zeros((2, 3, 4))
            assert a.flat.base is a

    def test_iter(self):
        for xp in (numpy, cupy):
            it = xp.zeros((2, 3, 4)).flat
            assert iter(it) is it

    def test_next(self):
        for xp in (numpy, cupy):
            a = testing.shaped_arange((2, 3, 4), xp)
            e = a.flatten()
            for ai, ei in zip(a.flat, e):
                assert ai == ei

    def test_len(self):
        for xp in (numpy, cupy):
            a = xp.zeros((2, 3, 4))
            assert len(a.flat) == 24
            assert len(a[::2].flat) == 12

    @testing.numpy_cupy_array_equal()
    def test_copy(self, xp):
        a = testing.shaped_arange((2, 3, 4), xp)
        o = a.flat.copy()
        assert a is not o
        return a.flat.copy()

    @testing.numpy_cupy_array_equal()
    def test_copy_next(self, xp):
        a = testing.shaped_arange((2, 3, 4), xp)
        it = a.flat
        it.__next__()
        return it.copy()  # Returns the flattened copy of whole `a`


@testing.parameterize(
    {'shape': (2, 3, 4), 'index': Ellipsis},
    {'shape': (2, 3, 4), 'index': 0},
    {'shape': (2, 3, 4), 'index': 10},
    {'shape': (2, 3, 4), 'index': slice(None)},
    {'shape': (2, 3, 4), 'index': slice(None, 10)},
    {'shape': (2, 3, 4), 'index': slice(None, None, 2)},
    {'shape': (2, 3, 4), 'index': slice(None, None, -1)},
    {'shape': (2, 3, 4), 'index': slice(10, None, -1)},
    {'shape': (2, 3, 4), 'index': slice(10, None, -2)},
    {'shape': (), 'index': slice(None)},
    {'shape': (10,), 'index': slice(None)},
)
class TestFlatiterSubscript(unittest.TestCase):

    @testing.for_CF_orders()
    @testing.for_all_dtypes()
    @testing.numpy_cupy_array_equal()
    def test_getitem(self, xp, dtype, order):
        a = testing.shaped_arange(self.shape, xp, dtype, order)
        return a.flat[self.index]

    @testing.for_CF_orders()
    @testing.for_all_dtypes()
    @testing.numpy_cupy_array_equal()
    def test_setitem_scalar(self, xp, dtype, order):
        a = xp.zeros(self.shape, dtype=dtype, order=order)
        a.flat[self.index] = 1
        return a

    @testing.for_CF_orders()
    @testing.for_all_dtypes()
    @testing.numpy_cupy_array_equal()
    def test_setitem_ndarray_1d(self, xp, dtype, order):
        if numpy.isscalar(self.index):
            pytest.skip()
        a = xp.zeros(self.shape, dtype=dtype, order=order)
        v = testing.shaped_arange((3,), xp, dtype, order)
        a.flat[self.index] = v
        return a

    @testing.for_CF_orders()
    @testing.for_all_dtypes()
    @testing.numpy_cupy_array_equal()
    def test_setitem_ndarray_nd(self, xp, dtype, order):
        if numpy.isscalar(self.index):
            pytest.skip()
        a = xp.zeros(self.shape, dtype=dtype, order=order)
        v = testing.shaped_arange((2, 3), xp, dtype, order)
        a.flat[self.index] = v
        return a

    @testing.with_requires('numpy<2.0')
    @testing.for_CF_orders()
    @testing.for_all_dtypes_combination(('a_dtype', 'v_dtype'))
    @testing.numpy_cupy_array_equal()
    def test_setitem_ndarray_different_types(
            self, xp, a_dtype, v_dtype, order):
        if numpy.isscalar(self.index):
            pytest.skip()
        a = xp.zeros(self.shape, dtype=a_dtype, order=order)
        v = testing.shaped_arange((3,), xp, v_dtype, order)
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', ComplexWarning)
            a.flat[self.index] = v
        return a


@testing.parameterize(
    {'shape': (2, 3, 4), 'index': None},
    {'shape': (2, 3, 4), 'index': (0,)},
    {'shape': (2, 3, 4), 'index': True},
    {'shape': (2, 3, 4), 'index': cupy.array([0])},
    {'shape': (2, 3, 4), 'index': [0]},
)
class TestFlatiterSubscriptIndexError(unittest.TestCase):

    @testing.for_all_dtypes()
    def test_getitem(self, dtype):
        a = testing.shaped_arange(self.shape, cupy, dtype)
        with pytest.raises(IndexError):
            a.flat[self.index]

    @testing.for_all_dtypes()
    def test_setitem(self, dtype):
        a = testing.shaped_arange(self.shape, cupy, dtype)
        v = testing.shaped_arange((1,), cupy, dtype)
        with pytest.raises(IndexError):
            a.flat[self.index] = v



import unittest

import pytest

import cupy
from cupy import testing


class TestRational(unittest.TestCase):

    @testing.for_dtypes(['?', 'e', 'f', 'd', 'F', 'D'])
    def test_gcd_dtype_check(self, dtype):
        a = cupy.random.randint(-10, 10, size=(10, 10)).astype(dtype)
        b = cupy.random.randint(-10, 10, size=(10, 10)).astype(dtype)
        with pytest.raises(TypeError):
            cupy.gcd(a, b)

    @testing.for_int_dtypes(no_bool=True)
    @testing.numpy_cupy_array_equal()
    def test_gcd_check_boundary_cases(self, xp, dtype):
        a = xp.array([0, -10, -5, 10, 410, 1, 6, 33])
        b = xp.array([0, 5, -10, -5, 20, 51, 6, 42])
        return xp.gcd(a, b)

    @testing.for_dtypes(['?', 'e', 'f', 'd', 'F', 'D'])
    def test_lcm_dtype_check(self, dtype):
        a = cupy.random.randint(-10, 10, size=(10, 10)).astype(dtype)
        b = cupy.random.randint(-10, 10, size=(10, 10)).astype(dtype)
        with pytest.raises(TypeError):
            cupy.lcm(a, b)

    @testing.for_int_dtypes(no_bool=True)
    @testing.numpy_cupy_array_equal()
    def test_lcm_check_boundary_cases(self, xp, dtype):
        a = xp.array([0, -10, -5, 10, 410, 1, 6, 33])
        b = xp.array([0, 5, -10, -5, 20, 51, 6, 42])
        return xp.lcm(a, b)



# flake8: NOQA
from cupyx.scipy.linalg._special_matrices import (
    tri, tril, triu, toeplitz, circulant, hankel,
    hadamard, leslie, kron, block_diag, companion,
    helmert, hilbert, dft,
    fiedler, fiedler_companion, convolution_matrix
)
from cupyx.scipy.linalg._solve_triangular import solve_triangular  # NOQA
from cupyx.scipy.linalg._decomp_lu import lu, lu_factor, lu_solve  # NOQA

# uarray backend support (NEP 31)
# The uarray feature for scipy.linalg is experimental.
# The interface can change in the future.
from cupyx.scipy.linalg._uarray import __ua_convert__  # NOQA
from cupyx.scipy.linalg._uarray import __ua_domain__  # NOQA
from cupyx.scipy.linalg._uarray import __ua_function__  # NOQA

from cupyx.scipy.linalg._array_utils import bandwidth  # NOQA
from cupyx.scipy.linalg._matfuncs import khatri_rao  # NOQA

from cupyx.scipy.linalg._matfuncs import expm  # NOQA
